package org.processmining.plugins.ppm;

import java.util.Date;
import java.util.Iterator;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeBooleanImpl;
import org.deckfour.xes.model.impl.XAttributeContinuousImpl;
import org.deckfour.xes.model.impl.XAttributeDiscreteImpl;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.deckfour.xes.model.impl.XAttributeTimestampImpl;
import org.deckfour.xes.model.impl.XEventImpl;
import org.deckfour.xes.model.impl.XLogImpl;
import org.deckfour.xes.model.impl.XTraceImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

@Plugin(name = "Convert Cheetah Log",
        parameterLabels = { "Log to be converted" },
        returnLabels = { "Converted log" },
        returnTypes = { XLog.class },
        userAccessible = true,
        help = "Converts a log from the experimantal Cheetah platform to a proper ProM event log")
public class ConvertCheetahLog 
{
	private static final boolean ADD_START_ACTIVITIES = false; //not only completed
	private static final boolean MORE_DETAILED_ACTIVITIES = true;
	private PluginContext context;
	

	@UITopiaVariant(affiliation = "Ghent University", 
			        author = "J. Claes", 
			        email = "jan.claes@ugent.be",
			        website = "http://www.janclaes.info")
	@PluginVariant(variantLabel = "Convert Cheetah Log, default", requiredParameterLabels = { 0 })
	public XLog convert(PluginContext context, XLog oldLog) {
		this.context = context;
		
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(oldLog.size());
		context.getProgress().setCaption("Converting log...");
		context.getProgress().setIndeterminate(false);
		
		XLog newLog = convertLog(oldLog);
		
		context.getFutureResult(0).setLabel("Converted from " + XConceptExtension.instance().extractName(oldLog));
		return newLog;
	}
	
	@UITopiaVariant(affiliation = "Ghent University", 
	        author = "J. Claes", 
	        email = "jan.claes@ugent.be",
	        website = "http://www.janclaes.info")
	@PluginVariant(variantLabel = "Convert Cheetah Log, default", requiredParameterLabels = { 0 })
	public XLog convert(UIPluginContext context, XLog oldLog) {
		this.context = context;
		
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(oldLog.size());
		context.getProgress().setCaption("Converting log...");
		context.getProgress().setIndeterminate(false);
		
		XLog newLog = convertLog(oldLog);
		
		context.getFutureResult(0).setLabel("Converted from " + XConceptExtension.instance().extractName(oldLog));
		return newLog;
	}
	
	
	private XLog convertLog(XLog oldLog) {
		XLog newLog = new XLogImpl(XLogFunctions.copyAttMap(oldLog.getAttributes()));
		for (int i=0; i<oldLog.size(); i++) {
			XTrace oldTrace = oldLog.get(i);
			XTrace newTrace = new XTraceImpl(XLogFunctions.copyAttMap(oldTrace.getAttributes()));
			for (int j=0; j<oldTrace.size(); j++) {
				XEvent oldEvent = oldTrace.get(j);
				addEvent(newTrace,oldEvent);
			}
			newLog.add(newTrace);
			context.getProgress().inc();
		}
		return newLog;
	}
	private void addEvent(XTrace newTrace, XEvent oldEvent) {
		XAttributeMap oldAttMap = oldEvent.getAttributes();
		XAttributeMap newAttMap = new XAttributeMapImpl();
		Iterator<XAttribute> attit = oldAttMap.values().iterator();
		while (attit.hasNext()) {
			XAttribute att = attit.next();
			String key = att.getKey();
			if (key.equals("concept:name"))
				key = "activityname";
			else if (key.equals("lifecycle:transition"))
				key = "concept:name";
			else if (ADD_START_ACTIVITIES && key.equals("add_node_start_time")) {
				addStartEvent(newTrace, oldEvent);
				continue;
			}
			newAttMap.put(key, clone(att,key));
		}
		String key = "lifecycle:transition";
		XAttribute att = new XAttributeLiteralImpl(key, "complete");
		newAttMap.put(key, att);
		
		if (MORE_DETAILED_ACTIVITIES) {
			String activity = newAttMap.get("concept:name").toString().toUpperCase(); 
			if (activity.equals("CREATE_NODE") || activity.equals("MOVE_NODE") || activity.equals("DELETE_NODE")) {
				if (newAttMap.containsKey("descriptor")) {
					String descriptor = newAttMap.get("descriptor").toString().toUpperCase();
					String newActivity = activity.replace("NODE", descriptor.replace("BPMN.", ""));
					XAttribute att2 = new XAttributeLiteralImpl("concept:name", newActivity);
					newAttMap.put("concept:name", att2);
				}
			} //CREATE_ACTIVITY, CREATE_AND, CREATE_XOR, CREATE_START_EVENT, CREATE_END_EVENT (same for MOVE, DELETE)
			else if (activity.equals("RENAME")) {
				String newActivity = activity;
				if (newAttMap.containsKey("descriptor")) {
					String descriptor = newAttMap.get("descriptor").toString().toUpperCase();
					newActivity = activity + "_" + descriptor.replace("BPMN.", "");
					newActivity = newActivity.replace("SEQUENCE_FLOW", "EDGE");
				}
				if (newAttMap.containsKey("new_name")) {
					if (!newAttMap.containsKey("name") || newAttMap.get("name").equals("") || newAttMap.get("name").equals("Sequence Flow"))
						newActivity = newActivity.substring(2);
				}
				XAttribute att2 = new XAttributeLiteralImpl("concept:name", newActivity);
				newAttMap.put("concept:name", att2);
			} //NAME_ACTIVITY, NAME_EDGE, RENAME_ACTIVITY, RENAME_EDGE
		}
		
		XEvent newEvent = new XEventImpl(newAttMap);
		newTrace.add(newEvent);
	}
	private void addStartEvent(XTrace newTrace, XEvent oldEvent) {
		XAttributeMap oldAttMap = oldEvent.getAttributes();
		XAttributeMap newAttMap = new XAttributeMapImpl();
		Iterator<XAttribute> attit = oldAttMap.values().iterator();
		while (attit.hasNext()) {
			XAttribute att = attit.next();
			String key = att.getKey();
			if (key.equals("concept:name"))
				key = "activityname";
			else if (key.equals("lifecycle:transition"))
				key = "concept:name";
			else if (key.equals("time:timestamp"))
				continue;
			else if (key.equals("add_node_start_time")) {
				key = "time:timestamp";
				String attValue = ((XAttributeLiteralImpl)att).getValue();
				att = new XAttributeTimestampImpl(key, new Date(Long.parseLong(attValue)));
				newAttMap.put(key, att);
				continue;
			}
			newAttMap.put(key, clone(att,key));
		}
		String key = "lifecycle:transition";
		XAttribute att = new XAttributeLiteralImpl(key, "start");
		newAttMap.put(key, att);
		XEvent newEvent = new XEventImpl(newAttMap);
		newTrace.add(newEvent);
	}
	private XAttribute clone(XAttribute att, String key) {
		if (XAttributeLiteralImpl.class.isInstance(att))			
			att = new XAttributeLiteralImpl(key, ((XAttributeLiteralImpl)att).getValue());
		else if (XAttributeTimestampImpl.class.isInstance(att))
			att = new XAttributeTimestampImpl(key, ((XAttributeTimestampImpl)att).getValue());
		else if (XAttributeBooleanImpl.class.isInstance(att))
			att = new XAttributeBooleanImpl(key, ((XAttributeBooleanImpl)att).getValue());
		else if (XAttributeContinuousImpl.class.isInstance(att))
			att = new XAttributeContinuousImpl(key, ((XAttributeContinuousImpl)att).getValue());
		else if (XAttributeDiscreteImpl.class.isInstance(att))
			att = new XAttributeDiscreteImpl(key, ((XAttributeDiscreteImpl)att).getValue());
		else
			att = (XAttribute)att.clone();
		return att;
	}
}
