package org.processmining.plugins.ppm;

import java.util.ArrayList;
import java.util.Date;

import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.deckfour.xes.model.impl.XEventImpl;
import org.deckfour.xes.model.impl.XLogImpl;
import org.deckfour.xes.model.impl.XTraceImpl;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class DerivePhases 
{
	private static final String MODELING = "MODELING";
	private static final String RECONCILIATION = "RECONCILIATION";
	private static final String COMPREHENSION = "COMPREHENSION";
	private static final long TRESHOLD_C = 30000;
	private static final long TRESHOLD_D = 2000;
	private static final long TRESHOLD_A = 4000;
	private PluginContext context;
	
	/*@Plugin(name = "Derive phases from Cheetah log",
			parameterLabels = { "Log to be altered" },
			returnLabels = { "Altered log" },
			returnTypes = { XLog.class },
			userAccessible = true,
			help = "Change events into phases in a log from the experimantal Cheetah platform")
	@UITopiaVariant(affiliation = "Ghent University", 
			        author = "J. Claes", 
			        email = "jan.claes@ugent.be",
			        website = "http://www.janclaes.info")
	public XLog convert(PluginContext context, XLog oldLog) {
		this.context = context;
		
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(oldLog.size());
		context.getProgress().setCaption("Altering log...");
		context.getProgress().setIndeterminate(false);
		
		XLog newLog = convertLog(oldLog);
		
		context.getFutureResult(0).setLabel("Altered from " + XConceptExtension.instance().extractName(oldLog));
		return newLog;
	}*/
	private XLog convertLog(XLog oldLog) {
		XLog newLog = new XLogImpl(XLogFunctions.copyAttMap(oldLog.getAttributes()));
		for (int i=0; i<oldLog.size(); i++) {
			XTrace oldTrace = oldLog.get(i);
			XTrace newTrace = new XTraceImpl(XLogFunctions.copyAttMap(oldTrace.getAttributes()));
			ArrayList<Phase> phases = new ArrayList<Phase>();
			Date currTime = new Date(), prevTime;
			Phase lastPhase = null;
			
			//detect phases
			for (int j=0; j<oldTrace.size(); j++) {
				XEvent oldEvent = oldTrace.get(j);
				prevTime = currTime;
				currTime = XLogFunctions.getTime(oldEvent);
				if (j==0) {
					lastPhase = new Phase(getPhase(oldEvent),currTime,currTime);
					phases.add(lastPhase);
					continue;
				}
				if (   (    !XLogFunctions.getTransition(oldTrace.get(j-1)).equals("start") 
					     || !XLogFunctions.getTransition(oldEvent).equals("complete")
					   ) 
				    && currTime.getTime()-prevTime.getTime() > TRESHOLD_C
				   ) {
					lastPhase = new Phase(COMPREHENSION,prevTime,currTime);
					phases.add(lastPhase);
				}
				if (getPhase(oldEvent).equals(lastPhase.getName()))
					lastPhase.setEnd(currTime);
				else {
					lastPhase = new Phase(getPhase(oldEvent),currTime,currTime);
					phases.add(lastPhase);;
				}
			}
			
			//merge phases 
			//if phase between two equals other phases has duration 
			// shorter then TRESHOLD_D (for modeling or reonciliation) or TRESHOLD_A (for comprehension))
			for (int j=2; j<phases.size(); j++) {
				Phase phase2 = phases.get(j);
				Phase phase1 = phases.get(j-1);
				Phase phase0 = phases.get(j-2);
				if (phase2.getName().equals(phase0.getName())) {
					long duration = phase1.getEnd().getTime() - phase1.getStart().getTime();
					if (    (phase0.getName().equals(COMPREHENSION) && duration < TRESHOLD_A) 
					     || (!phase0.getName().equals(COMPREHENSION) && duration < TRESHOLD_D) 
					    ) {
						phase0.setEnd(phase2.getEnd());
						phases.remove(phase1);
						phases.remove(phase2);
						j = (j-2 < 2)? 2 : j-2;						
					}
				}
			}
			
			//make new trace
			for (int j=0; j<phases.size(); j++) {
				Phase phase = phases.get(j);
				XEvent newEvent = makeEvent(phase.getName(),phase.getStart(),"start");
				newTrace.add(newEvent);
				newEvent = makeEvent(phase.getName(),phase.getEnd(),"complete");
				newTrace.add(newEvent);
			}
			newLog.add(newTrace);
			context.getProgress().inc();
		}
		return newLog;
	}
	private String getPhase(XEvent event) {
		String name = XLogFunctions.getName(event);
		if (name.toUpperCase().equals("CREATE_NODE"))
			return MODELING;
		else if (name.toUpperCase().equals("CREATE_EDGE"))
			return MODELING;
		else if (name.toUpperCase().equals("DELETE_NODE"))
			return MODELING;
		else if (name.toUpperCase().equals("DELETE_EDGE"))
			return MODELING;
		else if (name.toUpperCase().equals("RECONNECT_EDGE"))
			return MODELING;
		else if (name.toUpperCase().equals("CREATE_EDGE_BENDPOINT"))
			return RECONCILIATION;
		else if (name.toUpperCase().equals("MOVE_EDGE_BENDPOINT"))
			return RECONCILIATION;
		else if (name.toUpperCase().equals("DELETE_EDGE_BENDPOINT"))
			return RECONCILIATION;
		else if (name.toUpperCase().equals("MOVE_NODE"))
			return RECONCILIATION;
		else if (name.toUpperCase().equals("MOVE_EDGE"))
			return RECONCILIATION;
		else if (name.toUpperCase().equals("MOVE_EDGE_LABEL"))
			return RECONCILIATION;
		else if (name.toUpperCase().equals("RENAME"))
			return RECONCILIATION;
		else
			return "OTHER";
	}
	private XEvent makeEvent(String name, Date time, String transition) {
		XAttributeMap attMap = new XAttributeMapImpl();
		XLogFunctions.putLiteral(attMap, "concept:name", name);
		XLogFunctions.putLiteral(attMap, "lifecycle:transition", transition);
		XLogFunctions.putLiteral(attMap, "org:resource", "artificial");
		XLogFunctions.putTimestamp(attMap, "time:timestamp", time);
		XEvent newEvent = new XEventImpl(attMap);
		return newEvent;
	}
}
