package org.processmining.plugins.ppm;

import java.io.File;
import java.util.HashMap;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.deckfour.xes.model.impl.XEventImpl;
import org.deckfour.xes.model.impl.XLogImpl;
import org.deckfour.xes.model.impl.XTraceImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.plugins.ppm.ppmchart.PPMChartAnalysis;
import org.processmining.plugins.ppm.ppmchart.model.PPMChartModel;
import org.processmining.plugins.ppm.ppmchart.ui.ChunkPanel;
import org.processmining.plugins.ppm.ppmchart.ui.FilterPanel;
import org.processmining.plugins.ppm.ppmchart.ui.PPMChartOptionPanel;
import org.processmining.plugins.ppm.ppmchart.ui.PPMChartPanel;
import org.processmining.plugins.simplelogoperations.XLogFunctions;


public class ExportPPMCharts 
{
	public static final int NUMBER_OF_LOGS = 30;
		
	@UITopiaVariant(affiliation = "Ghent University", 
			        author = "J. Claes", 
			        email = "jan.claes@ugent.be",
			        website = "http://www.janclaes.info")
	@Plugin(name = "Export PPMCharts (one per trace)",
	parameterLabels = { "Log to generate and export PPMCharts for" },
	returnLabels = { "Original log (and chart figures in files system)" },
	returnTypes = { XLog.class },
	userAccessible = true,
	help = "Splits a log from the experimantal Cheetah platform to a log per contained model (with a trace per model object) and generates a PPMChart for each model")
	public static XLog export(PluginContext context, XLog log, ExportPPMParametersPanel parameters) {
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(2*log.size());
		context.getProgress().setCaption("Splitting log...");
		context.getProgress().setIndeterminate(false);
		
		exportLog(context, log, parameters);
		return log;
	}
	@UITopiaVariant(affiliation = "Ghent University", 
	        author = "J. Claes", 
	        email = "jan.claes@ugent.be",
	        website = "http://www.janclaes.info")
	@Plugin(name = "Export PPMCharts (one per trace)",
	parameterLabels = { "Log to generate and export PPMCharts for" },
	returnLabels = { "Original log (and chart figures in files system)" },
	returnTypes = { XLog.class },
	userAccessible = true,
	help = "Splits a log from the experimantal Cheetah platform to a log per contained model (with a trace per model object) and generates a PPMChart for each model")
	public static XLog export(UIPluginContext context, XLog log) {
		
		ExportPPMParametersPanel parameters = new ExportPPMParametersPanel(context, log);
		InteractionResult interActionResult = context.showConfiguration("Set export parameters", parameters);
		if (interActionResult.equals(InteractionResult.CANCEL)) {
			context.getFutureResult(0).cancel(true);
			return log;
		}
		
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(2*log.size());
		context.getProgress().setCaption("Splitting log...");
		context.getProgress().setIndeterminate(false);
		
		exportLog(context, log, parameters);
		return log;
		
	}
	private static void exportLog(PluginContext context, XLog log, ExportPPMParametersPanel parameters) {
		XLog[] logs = convertLog(context, log);
		JFileChooser fc = new JFileChooser();
		fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		if (fc.showSaveDialog(null) == JFileChooser.APPROVE_OPTION) {
			File dir = fc.getCurrentDirectory();
			for (XLog partLog : logs) {
				try {
					PPMChartAnalysis ppmca = new PPMChartAnalysis(context, new PPMChartModel(context, partLog));
					copyParameters(parameters, ppmca);
					if (parameters.getTotalWidth().equals("10 min"))
						PPMChartModel.timeOffset = 10 * 60 * 1000L;
					else if (parameters.getTotalWidth().equals("20 min"))
						PPMChartModel.timeOffset = 20 * 60 * 1000L;
					else if (parameters.getTotalWidth().equals("30 min"))
						PPMChartModel.timeOffset = 30 * 60 * 1000L;
					else if (parameters.getTotalWidth().equals("1 hour"))
						PPMChartModel.timeOffset = 60 * 60 * 1000L;
					else if (parameters.getTotalWidth().equals("1,5 hours"))
						PPMChartModel.timeOffset = 90 * 60 * 1000L;
					else if (parameters.getTotalWidth().equals("2 hours"))
						PPMChartModel.timeOffset = 120 * 60 * 1000L;
					PPMChartPanel ppmcp = ppmca.getPPMChartPanel();
					File f = new File(dir, XLogFunctions.getID(partLog) + ".png");
					ppmcp.setBounds(0,  0,  720,  540);
					ppmcp.export(f);
				}
				catch (Exception ex) {
					ex.printStackTrace();
					JOptionPane.showMessageDialog(null, "Error trying to save results to file", "Error", JOptionPane.ERROR_MESSAGE);
				}
				context.getProgress().inc();
			}
		}
	}
	public static XLog[] convertLog(PluginContext context, XLog log) {
		XLog[] newLogs = new XLog[log.size()];
		for (int i=0; i<newLogs.length; i++) {
			newLogs[i] = null;
		}
		for (int i=0; i<log.size(); i++) {
			XTrace oldTrace = log.get(i);
			newLogs[i] = convertTrace(oldTrace);
			context.getProgress().inc();
		}
		return newLogs;
	}
	public static XLog convertTrace(XTrace trace) {
		XLog log = new XLogImpl(XLogFunctions.copyAttMap(trace.getAttributes()));
		HashMap<String,XTrace> newTraces = new HashMap<String,XTrace>(); 
		for (int j=0; j<trace.size(); j++) {
			XEvent oldEvent = trace.get(j);
			String id = "-";
			try {
				id = oldEvent.getAttributes().get("id").toString();
			} catch (Exception ex) {
				continue;
			}
			if (!newTraces.containsKey(id)) {
				XTrace newTrace = new XTraceImpl(XLogFunctions.copyAttMap(oldEvent.getAttributes()));
				XLogFunctions.setName(newTrace, id);
				newTraces.put(id,newTrace);
			}
			XEvent newEvent = new XEventImpl(XLogFunctions.copyAttMap(oldEvent.getAttributes()));
			newTraces.get(id).add(newEvent);
		}
		Object[] keys = newTraces.keySet().toArray();
		for (int j=0; j<keys.length; j++) {
			log.add(newTraces.get(keys[j]));
		}
		log.getAttributes().put("concept:name",new XAttributeLiteralImpl("concept:name", XLogFunctions.getName(trace)));
		return log;
	}
	private static void copyParameters(ExportPPMParametersPanel parameters, PPMChartAnalysis ppmca) {
		PPMChartOptionPanel ppmcop = ppmca.getPPMChartOptionPanel();
		ppmcop.setColorStandard(parameters.getColorStandard());
		ppmcop.setShapeStandard(parameters.getShapeStandard());
		ppmcop.setSortStandard(parameters.getSortStandard());
		ppmcop.setComponentType(parameters.getComponentType());
		ppmcop.setTotalWidth(parameters.getTotalWidth());
		ppmcop.setWidthDivider(parameters.getWidthDivider());
		ppmcop.setZoomSliderX(parameters.getZoomSliderX());
		ppmcop.setZoomSliderY(parameters.getZoomSliderY());
		ppmcop.setTimeOption(parameters.getTimeOption());
		ppmcop.setDescCheckBoxSelected(parameters.isDescCheckBoxSelected());
		
		FilterPanel fp = ppmca.getFilterPanel();
		fp.setCreaSelected(parameters.isCreaSelected());
		fp.setMoveSelected(parameters.isMoveSelected());
		fp.setDeleSelected(parameters.isDeleSelected());
		fp.setNameSelected(parameters.isNameSelected());
		fp.setActSelected(parameters.isActSelected());
		fp.setStaSelected(parameters.isStaSelected());
		fp.setEndSelected(parameters.isEndSelected());
		fp.setAndSelected(parameters.isAndSelected());
		fp.setXorSelected(parameters.isXorSelected());
		fp.setEdgSelected(parameters.isEdgSelected());
		fp.setAllItemsSelected(parameters.isAllItemsSelected());
		
		ChunkPanel cp = ppmca.getSettingPanel().getChunkPanel();
		cp.setLocalityChunks(parameters.getLocalityChunks());
		cp.setSimilarityChunks(parameters.getLocalityChunks());
		cp.setTimingChunks(parameters.getLocalityChunks());
	}
}
