package org.processmining.plugins.ppm;

import java.util.Date;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.deckfour.xes.model.impl.XEventImpl;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class InsertComprehension 
{
	private final long TRESHOLD_C = 30000;
	private PluginContext context;
	
	@Plugin(name = "Insert Comprehension phases in Cheetah log",
			parameterLabels = { "Log to be altered" },
			returnLabels = { "Altered log" },
			returnTypes = { XLog.class },
			userAccessible = true,
			help = "Insert comprehension phases between events when duration between events is above treshold")
	@UITopiaVariant(affiliation = "Ghent University", 
			        author = "J. Claes", 
			        email = "jan.claes@ugent.be",
			        website = "http://www.janclaes.info")
	public XLog insert(PluginContext context, XLog log) {
		this.context = context;
		
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(log.size());
		context.getProgress().setCaption("Altering log...");
		context.getProgress().setIndeterminate(false);
		
		insertComprehension(log);
		
		context.getFutureResult(0).setLabel("Altered from " + XConceptExtension.instance().extractName(log));
		return log;
	}
	private void insertComprehension(XLog log) {
		for (int i=0; i<log.size(); i++) {
			XTrace trace = log.get(i);
			XEvent prev, curr = trace.get(0);
			for (int j=1; j<trace.size(); j++) {
				prev = curr;
				curr = trace.get(j);
				String prevType = XLogFunctions.getTransition(prev); 
				String currType = XLogFunctions.getTransition(curr);
				Date start = XLogFunctions.getTime(prev);
				Date end = XLogFunctions.getTime(curr);
				if (   (!prevType.equals("start") || !currType.equals("complete")) 
					&& end.getTime()-start.getTime() > TRESHOLD_C) {
					trace.add(j, makeEvent(start,"start"));
					trace.add(j, makeEvent(end,"complete"));
					j += 2;
				}					
			}
			context.getProgress().inc();
		}		
	}
	private XEvent makeEvent(Date time, String transition) {
		XAttributeMap attMap = new XAttributeMapImpl();
		XLogFunctions.putLiteral(attMap, "concept:name", "COMPREHENSION");
		XLogFunctions.putLiteral(attMap, "lifecycle:transition", transition);
		XLogFunctions.putLiteral(attMap, "org:resource", "artificial");
		XLogFunctions.putTimestamp(attMap, "time:timestamp", time);
		XEvent newEvent = new XEventImpl(attMap);
		return newEvent;
	}
}
