package org.processmining.plugins.ppm;

import java.util.HashMap;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XEventImpl;
import org.deckfour.xes.model.impl.XLogImpl;
import org.deckfour.xes.model.impl.XTraceImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.plugins.simplelogoperations.XLogFunctions;


public class SplitIntoProcessPerModel 
{
	public static final int NUMBER_OF_LOGS = 30;
		
	@UITopiaVariant(affiliation = "Ghent University", 
			        author = "J. Claes", 
			        email = "jan.claes@ugent.be",
			        website = "http://www.janclaes.info")
	@Plugin(name = "Split Cheetah Log into Processes per Model (max " + SplitIntoProcessPerModel.NUMBER_OF_LOGS + ")",
	parameterLabels = { "Log to be split" },
	returnLabels = { "Splitted log 1", "Splitted log 2", "Splitted log 3", "Splitted log 4", "Splitted log 5", 
	                 "Splitted log 6", "Splitted log 7", "Splitted log 8", "Splitted log 9", "Splitted log 10", 
	                 "Splitted log 11", "Splitted log 12", "Splitted log 13", "Splitted log 14", "Splitted log 15", 
	                 "Splitted log 16", "Splitted log 17", "Splitted log 18", "Splitted log 19", "Splitted log 20", 
	                 "Splitted log 21", "Splitted log 22", "Splitted log 23", "Splitted log 24", "Splitted log 25", 
	                 "Splitted log 26", "Splitted log 27", "Splitted log 28", "Splitted log 29", "Splitted log 30" },
	returnTypes = { XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, 
	                XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, 
	                XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, 
	                XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, 
	                XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, XLog.class },
	userAccessible = true,
	help = "Splits a log from the experimantal Cheetah platform to a log per contained model (with a trace per model object), maximum " + SplitIntoProcessPerModel.NUMBER_OF_LOGS)
	public static Object[] convert(PluginContext context, XLog oldLog) {
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(oldLog.size());
		context.getProgress().setCaption("Splitting log...");
		context.getProgress().setIndeterminate(false);
		
		XLog[] newLogs = convertLog(context, oldLog);
		
		for (int i=0; i<Math.min(NUMBER_OF_LOGS,oldLog.size()); i++)
			context.getFutureResult(i).setLabel("Model " + XLogFunctions.getName(oldLog.get(i)) + " from " + XConceptExtension.instance().extractName(oldLog));
		return newLogs;
	}
	@UITopiaVariant(affiliation = "Ghent University", 
	        author = "J. Claes", 
	        email = "jan.claes@ugent.be",
	        website = "http://www.janclaes.info")
	@Plugin(name = "Split Cheetah Log into Processes per Model (max " + SplitIntoProcessPerModel.NUMBER_OF_LOGS + ")",
	parameterLabels = { "Log to be split" },
	returnLabels = { "Splitted log 1", "Splitted log 2", "Splitted log 3", "Splitted log 4", "Splitted log 5", 
	                 "Splitted log 6", "Splitted log 7", "Splitted log 8", "Splitted log 9", "Splitted log 10", 
	                 "Splitted log 11", "Splitted log 12", "Splitted log 13", "Splitted log 14", "Splitted log 15", 
	                 "Splitted log 16", "Splitted log 17", "Splitted log 18", "Splitted log 19", "Splitted log 20", 
	                 "Splitted log 21", "Splitted log 22", "Splitted log 23", "Splitted log 24", "Splitted log 25", 
	                 "Splitted log 26", "Splitted log 27", "Splitted log 28", "Splitted log 29", "Splitted log 30" },
	returnTypes = { XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, 
	                XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, 
	                XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, 
	                XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, 
	                XLog.class, XLog.class, XLog.class, XLog.class, XLog.class, XLog.class },
	userAccessible = true,
	help = "Splits a log from the experimantal Cheetah platform to a log per contained model (with a trace per model object), maximum " + SplitIntoProcessPerModel.NUMBER_OF_LOGS)
	public static Object[] convert(UIPluginContext context, XLog oldLog) {
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(oldLog.size());
		context.getProgress().setCaption("Splitting log...");
		context.getProgress().setIndeterminate(false);
		
		XLog[] newLogs = convertLog(context, oldLog);
		
		for (int i=0; i<Math.min(NUMBER_OF_LOGS,oldLog.size()); i++)
			context.getFutureResult(i).setLabel("Model " + XLogFunctions.getName(oldLog.get(i)) + " from " + XConceptExtension.instance().extractName(oldLog));
		return newLogs;
	}
	private static XLog[] convertLog(PluginContext context, XLog oldLog) {
		XLog[] newLogs = new XLog[NUMBER_OF_LOGS];
		for (int i=0; i<newLogs.length; i++) {
			newLogs[i] = null;
		}
		for (int i=0; i<Math.min(NUMBER_OF_LOGS,oldLog.size()); i++) {
			XTrace oldTrace = oldLog.get(i);
			newLogs[i] = new XLogImpl(XLogFunctions.copyAttMap(oldTrace.getAttributes()));
			HashMap<String,XTrace> newTraces = new HashMap<String,XTrace>(); 
			for (int j=0; j<oldTrace.size(); j++) {
				XEvent oldEvent = oldTrace.get(j);
				String id = "-";
				try {
					id = oldEvent.getAttributes().get("id").toString();
				} catch (Exception ex) {
					continue;
				}
				if (!newTraces.containsKey(id)) {
					XTrace newTrace = new XTraceImpl(XLogFunctions.copyAttMap(oldEvent.getAttributes()));
					XLogFunctions.setName(newTrace, id);
					newTraces.put(id,newTrace);
				}
				XEvent newEvent = new XEventImpl(XLogFunctions.copyAttMap(oldEvent.getAttributes()));
				newTraces.get(id).add(newEvent);
			}
			Object[] keys = newTraces.keySet().toArray();
			for (int j=0; j<keys.length; j++) {
				newLogs[i].add(newTraces.get(keys[j]));
			}
			context.getProgress().inc();
		}
		return newLogs;
	}
}
