package org.processmining.plugins.ppm.ppmchart;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;

import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.ScrollPaneConstants;

import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.plugins.ppm.ppmchart.model.PPMChartModel;
import org.processmining.plugins.ppm.ppmchart.ui.FilterPanel;
import org.processmining.plugins.ppm.ppmchart.ui.OverviewPanel;
import org.processmining.plugins.ppm.ppmchart.ui.PPMChartOptionPanel;
import org.processmining.plugins.ppm.ppmchart.ui.PPMChartPanel;
import org.processmining.plugins.ppm.ppmchart.ui.SettingPanel;

public class PPMChartAnalysis extends JPanel implements ActionListener {

	private static final long serialVersionUID = -2720292789575028504L;

	private final PPMChartModel ccModel;

	// GUI attributes
	private final JPanel centerPanel = new JPanel();
	private final JTabbedPane tabbedPane = new JTabbedPane();
	private JSplitPane sp = new JSplitPane();
	private JSplitPane split = new JSplitPane();
	private JSplitPane spom = new JSplitPane();
	private OverviewPanel oPanel = null;
	private FilterPanel fPanel = null;
	private PPMChartPanel ppmChartPanel = null;
	private PPMChartOptionPanel ppmChartOptionPanel = null;
	private JScrollPane chartPane;
	private SettingPanel settingPanel;

	public PPMChartAnalysis(PluginContext context, PPMChartModel model) {
		ccModel = model;
		analyse();
	}

	public void analyse() {
		init();
	}

	private void init() {

		int height = getHeight() - this.getInsets().bottom - this.getInsets().top - 50;
		int width = getWidth() - this.getInsets().left - this.getInsets().right - 200;

		// initialize
		settingPanel = new SettingPanel(this);
		ppmChartOptionPanel = new PPMChartOptionPanel(this);
		ppmChartPanel = new PPMChartPanel(this);
		ppmChartOptionPanel.registerGUIListener(ppmChartPanel);

		// initialize settings of taskMap
		chartPane = new JScrollPane(ppmChartPanel, ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
				ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
		chartPane.setAutoscrolls(true);

		// repaint after moving scrollbar
		AdjustmentListener adjListener = new AdjustmentListener() {
			public void adjustmentValueChanged(AdjustmentEvent e) {
				reDrawBoxOnOverview();
				ppmChartPanel.updateUI();
				ppmChartPanel.repaint();
			}
		};
		chartPane.getHorizontalScrollBar().addAdjustmentListener(adjListener);
		chartPane.getVerticalScrollBar().addAdjustmentListener(adjListener);
		chartPane.setPreferredSize(new Dimension(width - 200, height - 10));

		// initialize overview Panel;
		oPanel = new OverviewPanel(this);
		oPanel.setPreferredSize(new Dimension(200, 130));
		
		// initialize filter Panel;
		fPanel = new FilterPanel(ppmChartPanel, ccModel.getLogReader());
		fPanel.setPreferredSize(new Dimension(200, 130));

		// initialize Panel including overview and metrics Panel
		spom = new JSplitPane(JSplitPane.VERTICAL_SPLIT, oPanel, fPanel);
		spom.setOneTouchExpandable(false);
		spom.setDividerLocation(131);
		spom.setDividerSize(3);

		// initialize main split panel
		split = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, chartPane, spom);
		split.setDividerLocation(split.getSize().width - split.getInsets().right - split.getDividerSize() - 200);

		split.setOneTouchExpandable(true);
		split.setResizeWeight(1.0);
		split.setDividerSize(3);
		split.setOneTouchExpandable(true);

		// Add PPM Chart Panel and Setting Panel
		tabbedPane.add("PPMChart", split);
		tabbedPane.add("Settings", settingPanel);
		tabbedPane.setSelectedIndex(0);
		tabbedPane.addChangeListener(new javax.swing.event.ChangeListener() {
			public void stateChanged(javax.swing.event.ChangeEvent e) {
				switch (tabbedPane.getSelectedIndex()) {
					case 0 :
						//						getOverviewPanel().setDrawBox(true);
						//						ppmChartPanel.changeWid)thSort();
						break;
					case 1 :
						settingPanel.initSettingPanel();
						break;
					default :
						break;
				}
			}
		});

		centerPanel.setLayout(new BorderLayout());
		centerPanel.add(tabbedPane, BorderLayout.CENTER);
		centerPanel.setMinimumSize(new Dimension(width, height));
		centerPanel.setAlignmentX(LEFT_ALIGNMENT);
		sp = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, ppmChartOptionPanel.getOptionPanel(true), centerPanel);//westPanel, centerPanel);
		sp.setDividerLocation(165);
		sp.setDividerSize(3);
		sp.setOneTouchExpandable(true);
		setLayout(new BorderLayout());
		this.add(sp, BorderLayout.CENTER);

		// zoom on 30 minutes
		/*long currentTimespan = ccModel.getLogMaxValue()+3600000 - ccModel.getLogMinValue();
		long desiredTimespan = 30 * 60 * 1000;
		double zoomLevel = currentTimespan * 1.0 / desiredTimespan;
		ppmChartOptionPanel.setZoomSliderXValue(zoomLevel);*/
		
		validate();
		this.repaint();
	}

	// public methods for attributes
	public PPMChartModel getPPMChartModel() {
		return ccModel;
	}

	public PPMChartPanel getPPMChartPanel() {
		return ppmChartPanel;
	}

	public SettingPanel getSettingPanel() {
		return settingPanel;
	}

	public PPMChartOptionPanel getPPMChartOptionPanel() {
		return ppmChartOptionPanel;
	}
	
	public FilterPanel getFilterPanel() {
		return fPanel;
	}

	// methods for dealing with zoom
	public Dimension getViewportSize() {
		return chartPane.getViewport().getSize();
	}

	public int getVerticalPosition() {
		return chartPane.getViewport().getViewPosition().y;
	}

	public int getHorizontalPosition() {
		return chartPane.getViewport().getViewPosition().x;
	}

	public JScrollPane getScrollPane() {
		return chartPane;
	}

	public void setScrollBarPosition(Point aP) {
		chartPane.getViewport().validate();
		if (aP.x < 0) {
			aP.x = 0;
		}
		if (aP.y < 0) {
			aP.y = 0;
		}
		if ((aP.x + chartPane.getViewport().getSize().getWidth()) > ppmChartPanel.getWidth()) {
			aP.x = (int) (ppmChartPanel.getWidth() - chartPane.getViewport().getSize().getWidth());
		}
		if ((aP.y + chartPane.getViewport().getSize().getHeight()) > ppmChartPanel.getHeight()) {
			aP.y = (int) (ppmChartPanel.getHeight() - chartPane.getViewport().getSize().getHeight());
		}
		chartPane.getViewport().setViewPosition(aP);
	}

	public void actionPerformed(ActionEvent e) {
		ccModel.updateStatisticsPanel(ppmChartOptionPanel);
		oPanel.setDrawBox(true);
		oPanel.repaint();
		oPanel.revalidate();
		tabbedPane.setSelectedIndex(0);
	}

	public void reDrawBoxOnOverview() {
		oPanel.repaint();
		oPanel.revalidate();
	}

	@Plugin(name = "PPMChart Visualization", 
			parameterLabels = "PPMChart", 
			returnTypes = PPMChartAnalysis.class, 
			returnLabels = "PPMChart", 
			userAccessible = false, 
			help = "Visualizes the ppmchart")
	@Visualizer(name = "PPMChart Visualization")
	public static PPMChartAnalysis visualize(PluginContext context, PPMChartModel dcm) {
		return new PPMChartAnalysis(context, dcm);
	}
}
