package org.processmining.plugins.ppm.ppmchart.model;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.processmining.plugins.ppm.ppmchart.ui.PPMChartOptionPanel;
import org.processmining.plugins.ppm.ppmchart.ui.UIUtil;

public class StatisticsModel {

	private HashMap<String, HashMap<String, DescriptiveStatistics>> overallMap = null;

	// GUI properties
	private final static String TIMESORTLABEL = "Time sort (metrics):";
	private final String[] timeSorts = { "milli", "seconds", "minutes", "half hours", "hours", "days", "weeks", "months", "years" };
	private final long[] dividers = { 1, 1000, 60000, 1800000L, 3600000L, 86400000L, 604800000L, 2592000000L, 31536000000L };
	private long timeDivider = 1000;
	private final PPMChartModel ccModel;
	private String type;
	private String timeOp;
	protected long timeOffset = 0;//1*3600000L;
	private ArrayList<String> sortedTitleList;
	private final UIUtil uiUtil = new UIUtil();

	public StatisticsModel(PPMChartModel ccModel) {
		this.ccModel = ccModel;
		overallMap = new HashMap<String, HashMap<String, DescriptiveStatistics>>();
		timeDivider = dividers[4];
	}

	public void initStatistics() {
		initStatistic(PPMChartModel.TIME_ACTUAL, PPMChartModel.ST_INST);
		initStatistic(PPMChartModel.TIME_ACTUAL, PPMChartModel.ST_TASK);
		//initStatistic(PPMChartModel.TIME_ACTUAL, PPMChartModel.ST_ORIG);
		//initStatistic(PPMChartModel.TIME_ACTUAL, PPMChartModel.ST_EVEN);
		//		initStatistic(CheetahChartPanel.TIME_RELATIVE_TIME,CheetahChartModel.ST_INST);
		initStatistic(PPMChartModel.TIME_RELATIVE_TIME, PPMChartModel.ST_TASK);
		//initStatistic(PPMChartModel.TIME_RELATIVE_TIME, PPMChartModel.ST_ORIG);
		//initStatistic(PPMChartModel.TIME_RELATIVE_TIME, PPMChartModel.ST_EVEN);
		initStatistic(PPMChartModel.TIME_RELATIVE_RATIO, PPMChartModel.ST_INST);
		initStatistic(PPMChartModel.TIME_RELATIVE_RATIO, PPMChartModel.ST_TASK);
		//initStatistic(PPMChartModel.TIME_RELATIVE_RATIO, PPMChartModel.ST_ORIG);
		//initStatistic(PPMChartModel.TIME_RELATIVE_RATIO, PPMChartModel.ST_EVEN);
		//initStatistic(PPMChartModel.TIME_LOGICAL_RELATIVE, PPMChartModel.ST_INST); // make it but not use it
		//initStatistic(PPMChartModel.TIME_LOGICAL_RELATIVE, PPMChartModel.ST_TASK);
		//initStatistic(PPMChartModel.TIME_LOGICAL_RELATIVE, PPMChartModel.ST_ORIG);
		//initStatistic(PPMChartModel.TIME_LOGICAL_RELATIVE, PPMChartModel.ST_EVEN);
	}

	private void initStatistic(String timeOption, String componentType) {
		HashMap<String, DescriptiveStatistics> tempMap = new HashMap<String, DescriptiveStatistics>();
		overallMap.put(timeOption + componentType, tempMap);
		tempMap.put(PPMChartModel.STATISTICS_OVERALL, new DescriptiveStatistics());
		for (String str : ccModel.getItemArrayList(componentType)) {
			tempMap.put(str, new DescriptiveStatistics());
		}
	}

	public void addValue(String timeOption, String instanceId, long duration) {
		overallMap.get(timeOption + PPMChartModel.ST_INST).get(instanceId).addValue(duration);
	}

	public void addValueOverAll(String timeOption, long duration) {
		overallMap.get(timeOption + PPMChartModel.ST_INST).get(PPMChartModel.STATISTICS_OVERALL).addValue(
				duration);
	}

	public void addValuesForOthers() {
		addValue(PPMChartModel.TIME_ACTUAL + PPMChartModel.ST_TASK, ccModel.getTaskCode(), ccModel
				.getCurrentTime(), ccModel.getItemArrayList(PPMChartModel.ST_TASK));
//		addValue(PPMChartModel.TIME_ACTUAL + PPMChartModel.ST_ORIG, ccModel.getOrigiantorCode(), ccModel
//				.getCurrentTime(), ccModel.getItemArrayList(PPMChartModel.ST_ORIG));
//		addValue(PPMChartModel.TIME_ACTUAL + PPMChartModel.ST_EVEN, ccModel.getEventCode(), ccModel
//				.getCurrentTime(), ccModel.getItemArrayList(PPMChartModel.ST_EVEN));
		addValue(PPMChartModel.TIME_RELATIVE_TIME + PPMChartModel.ST_TASK, ccModel.getTaskCode(), ccModel
				.getRelativeTime(), ccModel.getItemArrayList(PPMChartModel.ST_TASK));
//		addValue(PPMChartModel.TIME_RELATIVE_TIME + PPMChartModel.ST_ORIG, ccModel.getOrigiantorCode(), ccModel
//				.getRelativeTime(), ccModel.getItemArrayList(PPMChartModel.ST_ORIG));
//		addValue(PPMChartModel.TIME_RELATIVE_TIME + PPMChartModel.ST_EVEN, ccModel.getEventCode(), ccModel
//				.getRelativeTime(), ccModel.getItemArrayList(PPMChartModel.ST_EVEN));
		addValue(PPMChartModel.TIME_RELATIVE_RATIO + PPMChartModel.ST_TASK, ccModel.getTaskCode(), ccModel
				.getRelativeRatio(), ccModel.getItemArrayList(PPMChartModel.ST_TASK));
//		addValue(PPMChartModel.TIME_RELATIVE_RATIO + PPMChartModel.ST_ORIG, ccModel.getOrigiantorCode(), ccModel
//				.getRelativeRatio(), ccModel.getItemArrayList(PPMChartModel.ST_ORIG));
//		addValue(PPMChartModel.TIME_RELATIVE_RATIO + PPMChartModel.ST_EVEN, ccModel.getEventCode(), ccModel
//				.getRelativeRatio(), ccModel.getItemArrayList(PPMChartModel.ST_EVEN));
//		addValue(PPMChartModel.TIME_LOGICAL_RELATIVE + PPMChartModel.ST_TASK, ccModel.getTaskCode(), ccModel
//				.getLogicalRelative(), ccModel.getItemArrayList(PPMChartModel.ST_TASK));
//		addValue(PPMChartModel.TIME_LOGICAL_RELATIVE + PPMChartModel.ST_ORIG, ccModel.getOrigiantorCode(),
//				ccModel.getLogicalRelative(), ccModel.getItemArrayList(PPMChartModel.ST_ORIG));
//		addValue(PPMChartModel.TIME_LOGICAL_RELATIVE + PPMChartModel.ST_EVEN, ccModel.getEventCode(), ccModel
//				.getLogicalRelative(), ccModel.getItemArrayList(PPMChartModel.ST_EVEN));
	}

	public void addValue(String componentType, int[] code, long[] times, ArrayList<String> codeNames) {
		for (int k = 0; k < codeNames.size(); k++) {
			ArrayList<Long> list = new ArrayList<Long>();
			for (int j = 0; j < code.length; j++) {
				if (code[j] == k) {
					list.add(times[j]);
				}
			}
			Object[] ia = list.toArray();
			Arrays.sort(ia);
			for (int j = 1; j < ia.length; j++) {
				overallMap.get(componentType).get(codeNames.get(k)).addValue((Long) ia[j] - (Long) ia[j - 1]);
			}
			if (ia.length > 1) {
				overallMap.get(componentType).get(PPMChartModel.STATISTICS_OVERALL).addValue(
						(Long) ia[ia.length - 1] - (Long) ia[0]);
			}
		}
	}

	public JPanel getPanel(PPMChartOptionPanel ccop) {
		return update(ccop);
	}

	public JPanel update(PPMChartOptionPanel ccop) {
		type = ccop.getComponentType();
		timeOp = ccop.getTimeOption();
		sortedTitleList = ccModel.getSortedMapModel().getSortedItemArrayList(ccop.getTimeOption(),
				ccop.getComponentType(), ccop.getSortStandard(), ccop.isDescCheckBoxSelected());
		return displayPerformanceMetrics();
	}

	public JPanel displayPerformanceMetrics() {
		HashMap<String, DescriptiveStatistics> aMap = null;
		if (timeOp.equals(PPMChartModel.TIME_RELATIVE_TIME) && type.equals(PPMChartModel.ST_INST)) {
			aMap = overallMap.get(PPMChartModel.TIME_ACTUAL + PPMChartModel.ST_INST);
		} else {
			aMap = overallMap.get(timeOp + type);
		}
		// initialize panel
		JPanel jPanel = new JPanel();
		initPanel(jPanel);

		int size = 0;
		long[] startTime = ccModel.getSortedMapModel().getStartTimes(timeOp, type, sortedTitleList);
		long[] endTime = ccModel.getSortedMapModel().getEndTimes(timeOp, type, sortedTitleList);

		// make the overall table
		AbstractTableModel otm = new OverallMetricTableModel();
		if (timeOp.equals(PPMChartModel.TIME_ACTUAL)) {
			jPanel.add(drawTable(otm, type, aMap.get(PPMChartModel.STATISTICS_OVERALL), ccModel.getLogMinValue(),
					ccModel.getLogMaxValue()));
		} else if (timeOp.equals(PPMChartModel.TIME_RELATIVE_TIME)) {
			jPanel.add(drawTable(otm, type, aMap.get(PPMChartModel.STATISTICS_OVERALL), 0, ccModel
					.getLogRelativeMaxValue()));
		} else if (timeOp.equals(PPMChartModel.TIME_RELATIVE_RATIO)) {
			jPanel.add(drawTable(otm, type, aMap.get(PPMChartModel.STATISTICS_OVERALL), 0, ccModel
					.getLogRelativeRatioMaxValue()));
//		} else if (timeOp.equals(PPMChartModel.TIME_LOGICAL_RELATIVE)) {
//			jPanel.add(drawTable(otm, type, aMap.get(PPMChartModel.STATISTICS_OVERALL), 0, ccModel
//					.getLogLogiclRelativeMaxValue()));
		}
		jPanel.add(Box.createRigidArea(new Dimension(5, 10)));

		// make statistics panel
		for (String str : sortedTitleList) {
			DescriptiveStatistics currentDS = aMap.get(str);
			otm = new OneMetricTableModel();

			jPanel.add(drawTable(otm, str, currentDS, startTime[size], endTime[size]));
			jPanel.add(Box.createRigidArea(new Dimension(5, 10)));
			size++;
		}
		//make sure the pattern performance information is displayed properly
		jPanel.setPreferredSize(new Dimension(200, 50 + 138 * (size + 1)));
		jPanel.revalidate();
		jPanel.repaint();
		return jPanel;
	}

	// convenient method /////////////////////////////////////////////////////////////
	private void initPanel(JPanel jPanel) {
		jPanel.removeAll();
		jPanel.setLayout(new BoxLayout(jPanel, BoxLayout.Y_AXIS));

		//add time option menu
		jPanel.add(Box.createRigidArea(new Dimension(5, 10)));
		JPanel menuPanel = new JPanel(new BorderLayout());
		menuPanel.setPreferredSize(new Dimension(160, 45));
		menuPanel.setMaximumSize(new Dimension(180, 45));
		JLabel label = new JLabel(TIMESORTLABEL);
		label.setAlignmentX(Component.LEFT_ALIGNMENT);
		menuPanel.add(label, BorderLayout.NORTH);
		JComboBox timeBox = new JComboBox(timeSorts);
		timeBox.setSelectedIndex(4);
		timeBox.setMaximumSize(new Dimension(160, 20));
		timeBox.setAlignmentX(Component.LEFT_ALIGNMENT);
		registerGUIListener(timeBox);
		menuPanel.add(Box.createRigidArea(new Dimension(5, 0)));
		menuPanel.add(timeBox, BorderLayout.CENTER);
		jPanel.add(menuPanel);
		jPanel.add(Box.createRigidArea(new Dimension(5, 10)));
	}

	private JPanel drawTable(AbstractTableModel otm, String str, DescriptiveStatistics currentDS, long start, long end) {
		DefaultTableCellRenderer dtcr2 = new DefaultTableCellRenderer();
		dtcr2.setBackground(new Color(235, 235, 235));
		JTable table = new JTable(otm);
		table.setPreferredSize(new Dimension(200, 55));
		table.setMaximumSize(new Dimension(200, 55));
		table.getColumnModel().getColumn(0).setPreferredWidth(70);
		table.getColumnModel().getColumn(0).setMaxWidth(100);
		table.getTableHeader().setFont(new Font("SansSerif", Font.PLAIN, 12));
		table.getColumnModel().getColumn(0).setCellRenderer(dtcr2);
		table.setBorder(BorderFactory.createEtchedBorder());
		if (timeOp.equals(PPMChartModel.TIME_ACTUAL)) {
			table.setValueAt(DateFormat.getInstance().format(start - timeOffset), 0, 1);
			table.setValueAt(DateFormat.getInstance().format(end - timeOffset), 1, 1);
			table.setValueAt(uiUtil.formatString(currentDS.getMean() / timeDivider, 5), 2, 1);
			table.setValueAt(uiUtil.formatString(currentDS.getMin() / timeDivider, 5), 3, 1);
			table.setValueAt(uiUtil.formatString(currentDS.getMax() / timeDivider, 5), 4, 1);
		} else if (timeOp.equals(PPMChartModel.TIME_RELATIVE_TIME)) {
			table.setValueAt(uiUtil.formatDate(start), 0, 1);
			table.setValueAt(uiUtil.formatDate(end), 1, 1);
			table.setValueAt(uiUtil.formatString(currentDS.getMean() / timeDivider, 5), 2, 1);
			table.setValueAt(uiUtil.formatString(currentDS.getMin() / timeDivider, 5), 3, 1);
			table.setValueAt(uiUtil.formatString(currentDS.getMax() / timeDivider, 5), 4, 1);
		} else if (timeOp.equals(PPMChartModel.TIME_RELATIVE_RATIO)) {
			table.setValueAt(uiUtil.formatRatio(start), 0, 1);
			table.setValueAt(uiUtil.formatRatio(end), 1, 1);
			table.setValueAt(uiUtil.formatString(currentDS.getMean() / timeDivider, 5), 2, 1);
			table.setValueAt(uiUtil.formatString(currentDS.getMin() / timeDivider, 5), 3, 1);
			table.setValueAt(uiUtil.formatString(currentDS.getMax() / timeDivider, 5), 4, 1);
//		} else if (timeOp.equals(PPMChartModel.TIME_LOGICAL_RELATIVE)) {
//			if (type.equals(PPMChartModel.ST_INST)) {
//				table.setValueAt(uiUtil.formatString(start, 5), 0, 1);
//				table.setValueAt(uiUtil.formatString(end, 5), 1, 1);
//				table.setValueAt(uiUtil.formatString(1, 5), 2, 1);
//				table.setValueAt(uiUtil.formatString(1, 5), 3, 1);
//				table.setValueAt(uiUtil.formatString(1, 5), 4, 1);
//			} else {
//				table.setValueAt(uiUtil.formatString(start, 5), 0, 1);
//				table.setValueAt(uiUtil.formatString(end, 5), 1, 1);
//				table.setValueAt(uiUtil.formatString(currentDS.getMean(), 5), 2, 1);
//				table.setValueAt(uiUtil.formatString(currentDS.getMin(), 5), 3, 1);
//				table.setValueAt(uiUtil.formatString(currentDS.getMax(), 5), 4, 1);
//			}
		}
		JPanel tempPanel0 = new JPanel(new BorderLayout());
		table.setAlignmentX(Component.CENTER_ALIGNMENT);
		tempPanel0.setPreferredSize(new Dimension(160, 98));
		tempPanel0.setMaximumSize(new Dimension(180, 98));
		tempPanel0.add(table.getTableHeader(), BorderLayout.NORTH);
		tempPanel0.add(table, BorderLayout.CENTER);
		JPanel tempPanel = new JPanel(new BorderLayout());
		JLabel patternLabel = null;

		if (otm instanceof OverallMetricTableModel) {
			patternLabel = new JLabel("Component Type: " + str);
		} else {
			patternLabel = new JLabel("Component ID: " + str);
		}

		patternLabel.setAlignmentX(Component.LEFT_ALIGNMENT);

		JLabel frequencyLabel = null;
		if (otm instanceof OverallMetricTableModel) {
			frequencyLabel = new JLabel("# of components: " + currentDS.getN());
		} else {
			frequencyLabel = new JLabel("# of dots: " + (currentDS.getN() + 1));
		}

		frequencyLabel.setFont(new Font("SansSerif", Font.PLAIN, 12));
		tempPanel.add(patternLabel, BorderLayout.NORTH);
		tempPanel.add(frequencyLabel, BorderLayout.CENTER);
		tempPanel.add(tempPanel0, BorderLayout.SOUTH);
		return tempPanel;
	}

	private void registerGUIListener(final JComboBox timeBox) {
		//timeBox listener: set timeSort to the sort that is selected in the box
		//and set timeDivider to the corresponding divider.
		timeBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				timeDivider = dividers[timeBox.getSelectedIndex()];
				displayPerformanceMetrics();
			}
		});

	}
}
