package org.processmining.plugins.ppm.ppmchart.ui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.BoxLayout;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JTextField;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.processmining.plugins.ppm.ppmchart.model.ChunkModel;
import org.processmining.plugins.ppm.ppmchart.model.PPMChartModel;

public class ChunkPanel extends JPanel implements ItemListener{

	private static final long serialVersionUID = -2720292789575028503L;
	
	protected PPMChartPanel ccPanel = null;
	protected ChunkModel chunkModel = null;
	
	private final int localityChunks = 1;
	private final int similarityChunks = 1;
	private final int timingChunks = 1;
	private final int blocks = 1;
	private final JSlider localitySlider = new JSlider(1,50,localityChunks);
	private final JSlider similaritySlider = new JSlider(1,50,similarityChunks);
	private final JSlider timingSlider = new JSlider(1,50,timingChunks);
	private final JSlider blockSlider = new JSlider(1,50,blocks);
	private final JTextField localityTextField = new JTextField("" + localityChunks, 3);
	private final JTextField similarityTextField = new JTextField("" + similarityChunks, 3);
	private final JTextField timingTextField = new JTextField("" + timingChunks, 3);
	private final JTextField blockTextField = new JTextField("" + blocks, 3);
	
	public ChunkPanel(PPMChartPanel ccPanel, PPMChartModel ccModel) {
		this.ccPanel = ccPanel;
		this.chunkModel = new ChunkModel(ccModel.getLogReader(), localityChunks - 1, similarityChunks - 1, timingChunks - 1, blocks - 1);
				
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		setAlignmentX(Component.LEFT_ALIGNMENT);
		
		addPanel(ChunkModel.LOCALITY, "Number of locality chunks:", localityTextField, localitySlider);
		addPanel(ChunkModel.SIMILARITY, "Number of similarity chunks:", similarityTextField, similaritySlider);
		addPanel(ChunkModel.TIMING, "Number of timing chunks:", timingTextField, timingSlider);
		addPanel(ChunkModel.BLOCKS, "Number of blocks:", blockTextField, blockSlider);
	}
	private void addPanel(int type, String label, JTextField text, JSlider slider) {
		JPanel panel = new JPanel();
		panel.setLayout(new BorderLayout());
		panel.add(new JLabel(label), BorderLayout.CENTER);
		panel.add(text, BorderLayout.EAST);
		panel.add(slider, BorderLayout.SOUTH);
		text.addActionListener(new TextListener(type, text, slider));
		slider.addChangeListener(new SliderListener(type, slider, text));
		this.add(panel);
	}

	public void itemStateChanged(ItemEvent e) {		
		ccPanel.initBufferedImage();
		ccPanel.revalidate();
		ccPanel.repaint();		
	}
	public int getLocalityChunks() { return localitySlider.getValue(); }
	public int getSimilarityChunks() { return similaritySlider.getValue(); }
	public int getTimingChunks() { return timingSlider.getValue(); }
	public int getBlocks() { return blockSlider.getValue(); }
	public void setLocalityChunks(int i) { localitySlider.setValue(i); }
	public void setSimilarityChunks(int i) { similaritySlider.setValue(i); }
	public void setTimingChunks(int i) { timingSlider.setValue(i); }
	public void setBlocks(int i) { blockSlider.setValue(i); }
	public ChunkModel getChunkModel() { return chunkModel; }

	class TextListener implements ActionListener {
		private int prevValue = 1;
		private int type;
		private JTextField text;
		private JSlider slider;
		
		public TextListener(int type, JTextField text, JSlider slider) { this.type = type; this.text = text; this.slider = slider; prevValue = getText(); }
		public void actionPerformed(ActionEvent e) {
			int val = getText();
			if (val != prevValue) {
				slider.setValue(val);
				prevValue = val;
				chunkModel.set(type, val-1);
			}
			else
				text.setText("" + prevValue);
		}
		public int getText() {
			try {
				int val = Integer.parseInt(text.getText());
				if (val < 1)
					return prevValue;
				else
					return val;
			}
			catch (Exception ex) {
				return prevValue;
			}
		}
	}
	class SliderListener implements ChangeListener {
		private int type;
		private JTextField text;
		private JSlider slider;
		
		public SliderListener(int type, JSlider slider, JTextField text) { this.type = type; this.slider = slider; this.text = text; }
		public void stateChanged(ChangeEvent e) {
			text.setText("" + slider.getValue());
			chunkModel.set(type, slider.getValue()-1);
		}
	}
}