package org.processmining.plugins.ppm.ppmchart.ui;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.ArrayList;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class FilterPanel extends JPanel implements ItemListener{

	private static final long serialVersionUID = -2720292789575028503L;
	
	protected PPMChartPanel ccPanel = null;
	
	private final JCheckBox actCheckBox = new JCheckBox("activities", false);
	private final JCheckBox staCheckBox = new JCheckBox("start events", false);
	private final JCheckBox endCheckBox = new JCheckBox("end events", false);
	private final JCheckBox andCheckBox = new JCheckBox("and gateways", false);
	private final JCheckBox xorCheckBox = new JCheckBox("xor gateways", false);
	private final JCheckBox edgCheckBox = new JCheckBox("edges", false);
	
	private final JCheckBox creaCheckBox = new JCheckBox("create", false);
	private final JCheckBox moveCheckBox = new JCheckBox("move", false);
	private final JCheckBox deleCheckBox = new JCheckBox("delete", false);
	private final JCheckBox nameCheckBox = new JCheckBox("(re)name", false);

	private ArrayList<String> actInstances;
	private ArrayList<String> staInstances;
	private ArrayList<String> endInstances;
	private ArrayList<String> andInstances;
	private ArrayList<String> xorInstances;
	private ArrayList<String> edgInstances;
	private ArrayList<String> creaInstances;
	private ArrayList<String> moveInstances;
	private ArrayList<String> deleInstances;
	private ArrayList<String> nameInstances;
	private ArrayList<String> creaTasks;
	private ArrayList<String> moveTasks;
	private ArrayList<String> deleTasks;
	private ArrayList<String> nameTasks;
	
	private final JCheckBox allItemsCheckBox = new JCheckBox("hide elements with ", false);
	
	public FilterPanel(PPMChartPanel ccPanel, XLog log) {
		this.ccPanel = ccPanel;
		initLists(log);
		
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		setAlignmentX(Component.LEFT_ALIGNMENT);
		
		add(new JLabel("Hide next model elements"));
		add(actCheckBox);
		add(staCheckBox);
		add(endCheckBox);
		add(andCheckBox);
		add(xorCheckBox);
		add(edgCheckBox);
		add(Box.createRigidArea(new Dimension(5, 10)));
		
		add(new JLabel("Hide next operations"));
		add(creaCheckBox);
		add(moveCheckBox);
		add(deleCheckBox);
		add(nameCheckBox);
		add(Box.createRigidArea(new Dimension(5, 10)));
		add(allItemsCheckBox);
		add(new JLabel("       these operations"));
				
		actCheckBox.addItemListener(this);
		staCheckBox.addItemListener(this);
		endCheckBox.addItemListener(this);
		andCheckBox.addItemListener(this);
		xorCheckBox.addItemListener(this);
		edgCheckBox.addItemListener(this);
		
		creaCheckBox.addItemListener(this);
		moveCheckBox.addItemListener(this);
		deleCheckBox.addItemListener(this);
		nameCheckBox.addItemListener(this);
		allItemsCheckBox.addItemListener(this);
	}
	protected void initLists(XLog log) {
		actInstances = new ArrayList<String>();
		staInstances = new ArrayList<String>();
		endInstances = new ArrayList<String>();
		andInstances = new ArrayList<String>();
		xorInstances = new ArrayList<String>();
		edgInstances = new ArrayList<String>();
		creaInstances = new ArrayList<String>();
		moveInstances = new ArrayList<String>();
		deleInstances = new ArrayList<String>();
		nameInstances = new ArrayList<String>();
		creaTasks = new ArrayList<String>();
		moveTasks = new ArrayList<String>();
		deleTasks = new ArrayList<String>();
		nameTasks = new ArrayList<String>();
		
		for (XTrace trace : log) {
			String id = XLogFunctions.getName(trace);
			for (XEvent event : trace) {
				String activity = XLogFunctions.getName(event);
				if (activity.endsWith("_BENDPOINT")) {
					moveTasks.add(activity);
					if (!moveInstances.contains(id)) moveInstances.add(id);
					if (!edgInstances.contains(id)) edgInstances.add(id);
				}
				else if (activity.equals("RECONNECT_EDGE")) {
					creaTasks.add(activity);
					deleTasks.add(activity);
					if (!creaInstances.contains(id)) creaInstances.add(id);
					if (!deleInstances.contains(id)) deleInstances.add(id);
				}
				else if (activity.startsWith("CREATE_")) {
					creaTasks.add(activity);
					if (!creaInstances.contains(id)) creaInstances.add(id);
				}
				else if (activity.startsWith("MOVE_")) {
					moveTasks.add(activity);
					if (!moveInstances.contains(id)) moveInstances.add(id);
				}
				else if (activity.startsWith("DELETE_")) {
					deleTasks.add(activity);
					if (!deleInstances.contains(id)) deleInstances.add(id);
				}
				else if (activity.contains("NAME_")) {
					nameTasks.add(activity);
					if (!nameInstances.contains(id)) nameInstances.add(id);
				}
				else {
					activity = "" + activity;
				}
				
				if (activity.endsWith("_ACTIVITY")) {
					if (!actInstances.contains(id)) actInstances.add(id);
				}
				else if (activity.endsWith("_START_EVENT")) {
					if (!staInstances.contains(id)) staInstances.add(id);
				}
				else if (activity.endsWith("_END_EVENT")) {
					if (!endInstances.contains(id)) endInstances.add(id);
				}
				else if (activity.endsWith("_AND")) {
					if (!andInstances.contains(id)) andInstances.add(id);
				}
				else if (activity.endsWith("_XOR")) {
					if (!xorInstances.contains(id)) xorInstances.add(id);
				}
				else if (activity.endsWith("_EDGE") || activity.equals("MOVE_EDGE_LABEL")) {
					if (!edgInstances.contains(id)) edgInstances.add(id);
				}
			}
		}
	}

	public void itemStateChanged(ItemEvent e) {		
		ccPanel.initBufferedImage();
		ccPanel.revalidate();
		ccPanel.repaint();		
	}
	public boolean hideTask(String task) {
		if (creaCheckBox.isSelected() && creaTasks.contains(task)) return true;
		if (moveCheckBox.isSelected() && moveTasks.contains(task)) return true;
		if (deleCheckBox.isSelected() && deleTasks.contains(task)) return true;
		if (nameCheckBox.isSelected() && nameTasks.contains(task)) return true;

		return false;
	}
	public boolean hideInstance(String instance) {
		if (actCheckBox.isSelected() && actInstances.contains(instance)) return true;
		if (staCheckBox.isSelected() && staInstances.contains(instance)) return true;
		if (endCheckBox.isSelected() && endInstances.contains(instance)) return true;
		if (andCheckBox.isSelected() && andInstances.contains(instance)) return true;
		if (xorCheckBox.isSelected() && xorInstances.contains(instance)) return true;
		if (edgCheckBox.isSelected() && edgInstances.contains(instance)) return true;

		if (allItemsCheckBox.isSelected()) {
			if (creaCheckBox.isSelected() && creaInstances.contains(instance)) return true;
			if (moveCheckBox.isSelected() && moveInstances.contains(instance)) return true;
			if (deleCheckBox.isSelected() && deleInstances.contains(instance)) return true;
			if (nameCheckBox.isSelected() && nameInstances.contains(instance)) return true;
		}
		
		return false;
	}
	
	public boolean isCreaSelected() { return creaCheckBox.isSelected(); }
	public boolean isMoveSelected() { return moveCheckBox.isSelected(); }
	public boolean isDeleSelected() { return deleCheckBox.isSelected(); }
	public boolean isNameSelected() { return nameCheckBox.isSelected(); }

	public boolean isActSelected() { return actCheckBox.isSelected(); }
	public boolean isStaSelected() { return staCheckBox.isSelected(); }
	public boolean isEndSelected() { return endCheckBox.isSelected(); }
	public boolean isAndSelected() { return andCheckBox.isSelected(); }
	public boolean isXorSelected() { return xorCheckBox.isSelected(); }
	public boolean isEdgSelected() { return edgCheckBox.isSelected(); }

	public boolean isAllItemsSelected() { return allItemsCheckBox.isSelected(); }
	
	public void setCreaSelected(boolean b) { creaCheckBox.setSelected(b); }
	public void setMoveSelected(boolean b) { moveCheckBox.setSelected(b); }
	public void setDeleSelected(boolean b) { deleCheckBox.setSelected(b); }
	public void setNameSelected(boolean b) { nameCheckBox.setSelected(b); }

	public void setActSelected(boolean b) { actCheckBox.setSelected(b); }
	public void setStaSelected(boolean b) { staCheckBox.setSelected(b); }
	public void setEndSelected(boolean b) { endCheckBox.setSelected(b); }
	public void setAndSelected(boolean b) { andCheckBox.setSelected(b); }
	public void setXorSelected(boolean b) { xorCheckBox.setSelected(b); }
	public void setEdgSelected(boolean b) { edgCheckBox.setSelected(b); }

	public void setAllItemsSelected(boolean b) { allItemsCheckBox.setSelected(b); }
}
