package org.processmining.plugins.ppm.ppmchart.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JColorChooser;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.filechooser.FileNameExtensionFilter;

import org.processmining.plugins.ppm.ppmchart.PPMChartAnalysis;
import org.processmining.plugins.ppm.ppmchart.model.PPMChartModel;

public class SettingPanel extends JPanel {

	private static final long serialVersionUID = -2720292789575028505L;

	protected JLabel colorsettingLabel = new JLabel("Change colors by pressing buttons");
	protected ColorReference colorReference;
	private JPanel colorPanel;

	protected ShapeReference shapeReference;
	private JPanel shapePanel;

	private final PPMChartAnalysis ppmChartAnalysis;
	private PPMChartPanel ppmChartPanel;
	private PPMChartModel ccModel;

	// for color reference
	private JLabel[] tempLable;
	private JButton[] tempButton;
	private JPanel colorChoosePanel;
	private JPanel colorMPanel;
	private JPanel colorMainPanel;
	private JScrollPane colorScrollPane;

	// for shape reference
	private JLabel[] tempLabel;
	private JComboBox[] tempCombo;
	private JPanel shapeChoosePanel;
	private JPanel shapeMPanel;
	private JPanel shapeMainPanel;
	private JScrollPane shapeScrollPane;
	
	// for chunking
	private ChunkPanel chunkPanel;

	//	private JTextField colorFile = new JTextField();
	//	private JButton chooseColorButton = new JButton();
	private JLabel fbackColorLabel;
	private JLabel sbackColorLabel;
	private JButton fbackColorButton = new JButton();
	private JButton sbackColorButton = new JButton();
	private JButton exportButton = new JButton();

	private Color colorLogDark = new Color(170, 170, 160);
	private Color colorLogBright = new Color(210, 210, 200);

	public SettingPanel(PPMChartAnalysis aPPMChartAnalysis) {
		ppmChartAnalysis = aPPMChartAnalysis;
		ppmChartPanel = aPPMChartAnalysis.getPPMChartPanel();
		ccModel = aPPMChartAnalysis.getPPMChartModel();
		chunkPanel = new ChunkPanel(ppmChartAnalysis.getPPMChartPanel(), ccModel);
	}

	public void initSettingPanel() {
		if (ppmChartPanel == null) {
			ppmChartPanel = ppmChartAnalysis.getPPMChartPanel();
			ccModel = ppmChartAnalysis.getPPMChartModel();
		}
		removeAll();
		// event panel
		JPanel p0 = new JPanel();
		p0.setLayout(new BoxLayout(p0, BoxLayout.PAGE_AXIS));
		p0.add(Box.createRigidArea(new Dimension(5, 10)));

		// init paint panel
		initColorPanel();
		initShapePanel();
		this.add(chunkPanel);
		
		exportButton.setText("Export");
		exportButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				JFileChooser fc = new JFileChooser();
				fc.setFileFilter(new FileNameExtensionFilter("*.png files", "png"));
				if (fc.showSaveDialog(null) == JFileChooser.APPROVE_OPTION) {
					try {
						ppmChartPanel.export(fc.getSelectedFile());
					}
					catch (Exception ex) {
						JOptionPane.showMessageDialog(null, "Error trying to save results to file", "Error", JOptionPane.ERROR_MESSAGE);
					}
				}
			}
		});
		add(exportButton);
	}

	// method dealing with color
	private void initColorPanel() {

		colorReference = ppmChartPanel.getColorReference();

		String type = ppmChartAnalysis.getPPMChartOptionPanel().getColorStandard();

		if (type.equals(PPMChartModel.STR_NONE)) {
			return;
		}
		ArrayList<String> keySet = ccModel.getItemArrayList(type);

		int rows = keySet.size() + 2;
		colorPanel = new JPanel(new GridLayout(rows, 1));
		colorPanel.setPreferredSize(new Dimension(300, rows * 15));
		colorPanel.setMaximumSize(new Dimension(300, rows * 15));

		tempLable = new JLabel[keySet.size()];
		tempButton = new JButton[keySet.size()];

		int i = 0;
		for (Iterator<String> itr = keySet.iterator(); itr.hasNext();) {
			String tempString = itr.next();
			tempLable[i] = new JLabel(tempString + ":");
			tempLable[i].setToolTipText(tempString);
			tempButton[i] = new JButton("push to change");
			tempButton[i].setForeground(colorReference.getColor(tempString));
			tempButton[i].setToolTipText(tempString);
			tempButton[i].setActionCommand(tempString);
			tempButton[i].addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					JButton tempButton = (JButton) e.getSource();
					Color newColor = JColorChooser.showDialog(tempButton, "Choose Background Color", tempButton
							.getBackground());
					if (newColor != null) {
						tempButton.setForeground(newColor);
						assignColor(tempButton.getActionCommand(), newColor);
					}
				}
			});
			colorChoosePanel = new JPanel(new GridLayout(1, 2));
			colorChoosePanel.add(tempLable[i]);
			colorChoosePanel.add(tempButton[i]);
			colorPanel.add(colorChoosePanel);
			i++;
		}
		
		sbackColorLabel = new JLabel("First Background");
		sbackColorLabel.setToolTipText("First Background");
		sbackColorButton = new JButton("push to change");
		sbackColorButton.setForeground(colorLogBright);
		sbackColorButton.setToolTipText("First Background");
		sbackColorButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				JButton tempButton = (JButton) e.getSource();
				Color newColor = JColorChooser.showDialog(tempButton, "Choose Background Color", tempButton
						.getBackground());
				if (newColor != null) {
					tempButton.setForeground(newColor);
					colorLogBright = newColor;
				}
			}
		});
		colorChoosePanel = new JPanel(new GridLayout(1, 2));
		colorChoosePanel.add(sbackColorLabel);
		colorChoosePanel.add(sbackColorButton);
		colorPanel.add(colorChoosePanel);
		
		fbackColorLabel = new JLabel("Second Background");
		fbackColorLabel.setToolTipText("Second Background");
		fbackColorButton = new JButton("push to change");
		fbackColorButton.setForeground(colorLogDark);
		fbackColorButton.setToolTipText("Second Background");
		fbackColorButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				JButton tempButton = (JButton) e.getSource();
				Color newColor = JColorChooser.showDialog(tempButton, "Choose Background Color", tempButton
						.getBackground());
				if (newColor != null) {
					tempButton.setForeground(newColor);
					colorLogDark = newColor;
				}
			}
		});
		colorChoosePanel = new JPanel(new GridLayout(1, 2));
		colorChoosePanel.add(fbackColorLabel);
		colorChoosePanel.add(fbackColorButton);
		colorPanel.add(colorChoosePanel);
		
		colorMPanel = null;
		colorMPanel = new JPanel();
		colorMPanel.setLayout(new BoxLayout(colorMPanel, BoxLayout.Y_AXIS));
		colorMPanel.add(colorPanel);
		colorScrollPane = new JScrollPane(colorMPanel);
		colorScrollPane.setPreferredSize(new Dimension(310, 410));
		colorMainPanel = new JPanel();
		colorMainPanel.setLayout(new BoxLayout(colorMainPanel, BoxLayout.Y_AXIS));
		JLabel changeColorLabel = new JLabel("           Set colors");
		colorMainPanel.add(changeColorLabel);
		colorMainPanel.add(colorScrollPane);

		this.add(colorMainPanel);
	}
	private void initShapePanel() {

		shapeReference = ppmChartPanel.getShapeReference();

		String type = ppmChartAnalysis.getPPMChartOptionPanel().getShapeStandard();

		if (type.equals(PPMChartModel.STR_NONE)) {
			return;
		}
		ArrayList<String> keySet = ccModel.getItemArrayList(type);

		shapePanel = new JPanel(new GridLayout(keySet.size(), 1));
		shapePanel.setPreferredSize(new Dimension(300, keySet.size() * 15));
		shapePanel.setMaximumSize(new Dimension(300, keySet.size() * 15));

		tempLabel = new JLabel[keySet.size()];
		tempCombo = new JComboBox[keySet.size()];

		int i = 0;
		for (Iterator<String> itr = keySet.iterator(); itr.hasNext();) {
			String tempString = itr.next();
			tempLabel[i] = new JLabel(tempString + ":");
			tempLabel[i].setToolTipText(tempString);
			tempCombo[i] = newComboBox(shapeReference.getShapeList(), shapeReference.getShape(tempString)); 
			tempCombo[i].setToolTipText(tempString);
			tempCombo[i].setActionCommand(tempString);
			tempCombo[i].addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					JComboBox tempCombo = (JComboBox) e.getSource();
					assignShape(tempCombo.getActionCommand(), shapeReference.getShapeList()[tempCombo.getSelectedIndex()]);
				}
			});
			shapeChoosePanel = new JPanel(new GridLayout(1, 2));
			shapeChoosePanel.add(tempLabel[i]);
			shapeChoosePanel.add(tempCombo[i]);
			shapePanel.add(shapeChoosePanel);
			i++;
		}
		shapeMPanel = null;
		shapeMPanel = new JPanel();
		shapeMPanel.setLayout(new BoxLayout(shapeMPanel, BoxLayout.Y_AXIS));
		shapeMPanel.add(shapePanel);
		shapeScrollPane = new JScrollPane(shapeMPanel);
		shapeScrollPane.setPreferredSize(new Dimension(310, 410));
		shapeMainPanel = new JPanel();
		shapeMainPanel.setLayout(new BoxLayout(shapeMainPanel, BoxLayout.Y_AXIS));
		JLabel changeShapeLabel = new JLabel("           Set shapes");
		shapeMainPanel.add(changeShapeLabel);
		shapeMainPanel.add(shapeScrollPane);

		this.add(shapeMainPanel);
	}
	private JComboBox newComboBox(String[] items, String selectedItem) {
		JComboBox combo = new JComboBox();
		for (int i=0; i<items.length; i++) {
			combo.addItem(items[i]);
			if (items[i].equals(selectedItem))
				combo.setSelectedIndex(combo.getItemCount() - 1);
		}
		return combo;
	}

	/*private void redrawColorPart() {
		String type = ppmChartAnalysis.getPPMChartOptionPanel().getColorStandard();
		colorMPanel.remove(colorPanel);
		this.remove(colorMainPanel);

		if (type.equals(PPMChartModel.STR_NONE)) {
			return;
		}
		ArrayList<String> keySet = ccModel.getItemArrayList(type);

		colorPanel = null;
		colorPanel = new JPanel(new GridLayout(keySet.size(), 1));
		colorPanel.setPreferredSize(new Dimension(250, keySet.size() * 15));
		colorPanel.setMaximumSize(new Dimension(250, keySet.size() * 15));

		int i = 0;
		for (Iterator<String> itr = keySet.iterator(); itr.hasNext();) {
			String tempString = itr.next();
			tempLable[i] = new JLabel(tempString + ":");
			tempLable[i].setToolTipText(tempString);
			tempButton[i] = new JButton("push to change");
			tempButton[i].setForeground(colorReference.getColor(tempString));
			tempButton[i].setToolTipText(tempString);
			tempButton[i].setActionCommand(tempString);
			tempButton[i].addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					JButton tempButton = (JButton) e.getSource();
					Color newColor = JColorChooser.showDialog(tempButton, "Choose Background Color", tempButton
							.getBackground());
					if (newColor != null) {
						tempButton.setBackground(newColor);
						assignColor(tempButton.getActionCommand(), newColor);
					}
				}
			});
			colorChoosePanel = new JPanel(new GridLayout(1, 2));
			colorChoosePanel.add(tempLable[i]);
			colorChoosePanel.add(tempButton[i]);
			colorPanel.add(colorChoosePanel);
			i++;
		}
		colorMPanel.add(colorPanel);
		colorMPanel = null;
		colorMPanel = new JPanel();
		colorMPanel.setLayout(new BoxLayout(colorMPanel, BoxLayout.Y_AXIS));
		colorMPanel.add(colorPanel);
		colorMPanel.repaint();
		colorMPanel.revalidate();
		colorScrollPane = new JScrollPane(colorMPanel);
		colorScrollPane.setPreferredSize(new Dimension(260, 410));
		colorMainPanel = null;
		colorMainPanel = new JPanel();
		colorMainPanel.setLayout(new BoxLayout(colorMainPanel, BoxLayout.Y_AXIS));
		JLabel changeColorLabel = new JLabel("           Set colors");
		colorMainPanel.add(changeColorLabel);
		colorMainPanel.add(colorScrollPane);

		JPanel tempPanel = new JPanel(new GridLayout(1, 2));
		tempPanel.add(sbackColorLabel);
		tempPanel.add(sbackColorButton);
		tempPanel.add(fbackColorLabel);
		tempPanel.add(fbackColorButton);
		colorMainPanel.add(tempPanel);

		//		colorMainPanel.add(colorFile);
		//		colorMainPanel.add(chooseColorButton);

		this.add(colorMainPanel);
		this.repaint();
		revalidate();
	}*/


	public Color getFBcolor() {
		return colorLogDark;
	}

	public Color getSBcolor() {
		return colorLogBright;
	}

	public void assignColor(String name, Color newColor) {
		colorReference.assignColor(name, newColor);
	}
	public void assignShape(String name, String newShape) {
		shapeReference.assignShape(name, newShape);
	}

	/*public void changeColorPanel() {
		if (colorMainPanel != null) {
			this.remove(colorMainPanel);
		}
		colorMainPanel = null;
		initColorPanel();
		this.repaint();
	}*/
	
	public ChunkPanel getChunkPanel() {
		return chunkPanel;
	}
}
