package org.processmining.plugins.ppm.statistics;

import java.util.*;

import org.deckfour.xes.model.*;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class BlockAnalyser {
	private ArrayList<String> nodes = new ArrayList<String>();
	private ArrayList<Edge> edges = new ArrayList<Edge>();
	private HashMap<String, ArrayList<String>> removedEdges = new HashMap<String, ArrayList<String>>();
	
	private HashMap<String, ArrayList<Path>> nodePath = new HashMap<String, ArrayList<Path>>();
	
	private ArrayList<Block> blocks = new ArrayList<Block>();
	private HashMap<String, ArrayList<Block>> nodeBlock = new HashMap<String, ArrayList<Block>>();
	
	private int numBlockAsAWhole, numBlocksInPieces;
	private int maxSimulBlock;
	
	public BlockAnalyser(XTrace trace) {
		for (XEvent event : trace) {
			String activity = XLogFunctions.getName(event).toUpperCase();
			if (!event.getAttributes().containsKey("id"))
				continue;
			String id = event.getAttributes().get("id").toString();
			if (    activity.equals("CREATE_ACTIVITY")
			     || activity.equals("CREATE_START_EVENT")
			     || activity.equals("CREATE_END_EVENT")
			     || activity.equals("CREATE_XOR")
			     || activity.equals("CREATE_AND"))
				addNode(id);
			else if (activity.equals("CREATE_EDGE")) {
				String source = event.getAttributes().get("source_node_id").toString();
				String target = event.getAttributes().get("target_node_id").toString();
				addEdge(id, source, target);
			}
			else if (    activity.equals("DELETE_ACTIVITY")
			          || activity.equals("DELETE_START_EVENT")
			          || activity.equals("DELETE_END_EVENT")
			          || activity.equals("DELETE_XOR") 
			          || activity.equals("DELETE_AND"))
				removeNode(id);
			else if (activity.equals("DELETE_EDGE")) {
				removeEdge(id);
			}
			else if (activity.equals("RECONNECT_EDGE")) {
				try {
					String source = event.getAttributes().get("source_node_id").toString();
					String target = event.getAttributes().get("target_node_id").toString();
					//TODO
					removeEdge(id);
					addEdge(id, source, target);
				}
				catch(Exception ex) { //ignore			
				}
			}
		}
		nodePath = null;
		analyse(trace);
		nodeBlock = null;
	}
	
	public void addNode(String node) {
		nodes.add(node);
		nodePath.put(node, new ArrayList<Path>());
	}
	public void removeNode(String node) {
		nodes.remove(node);
		nodePath.put(node, new ArrayList<Path>()); //TODO
	}
	public void addEdge(String id, String source, String target){
		Edge edge = findEdge(id);
		if (edge != null)
			edge = null;//TODO
		if (removedEdges.containsKey(source) && removedEdges.get(source).contains(target))
			return;
		edges.add(new Edge(id, source, target));
		addMergeOrExtendPath(source, target);
		checkBlocks(source);
		checkBlocks(target);
	}
	public void removeEdge(String id){
		Edge edge = findEdge(id);
		String source = edge.getSource();
		String target = edge.getTarget();
		if (edge == null)
			edge = null;//TODO
		if (!removedEdges.containsKey(source))
			removedEdges.put(source, new ArrayList<String>());
		else if (!removedEdges.get(source).contains(target))
				removedEdges.get(source).add(target);
	}
	private Edge findEdge(String id) {
		for (Edge edge : edges)
			if (edge.getId().equals(id))
				return edge;
		return null;
	}
	private void splitExistingPathsIfNewEdgeStartsOrEndsInTheMiddleOfThatPath(String source, String target) {
		if (    nodePath.get(source).size() == 1 
		     && !nodePath.get(source).get(0).peekFirst().equals(source)
		     && !nodePath.get(source).get(0).peekLast().equals(source)) {
			Path p = nodePath.get(source).get(0);
			Path p1 = nodePath.get(source).get(0).split(source);
			for (String node : p1) {
				nodePath.get(node).remove(p);
				nodePath.get(node).add(p1);
			}
			nodePath.get(source).add(p);
		}
		if (    nodePath.get(target).size() == 1 
		     && !nodePath.get(target).get(0).peekFirst().equals(target)
		     && !nodePath.get(target).get(0).peekLast().equals(target)) {
			Path p = nodePath.get(target).get(0);
			Path p1 = nodePath.get(target).get(0).split(target);
			for (String node : p1) {
				nodePath.get(node).remove(p);
				nodePath.get(node).add(p1);
			}
			nodePath.get(target).add(p);
		}
	}
	private void addMergeOrExtendPath(String source, String target) {
		splitExistingPathsIfNewEdgeStartsOrEndsInTheMiddleOfThatPath(source, target);
		if (nodePath.get(source).size() != 1 && nodePath.get(target).size() != 1) {
			//add
			Path p = new Path(source, target);
			nodePath.get(source).add(p);
			nodePath.get(target).add(p);
		}
		else if (    nodePath.get(source).size() == 1 
				  && nodePath.get(source).get(0).peekLast().equals(source)
				  && nodePath.get(target).size() == 1
				  && nodePath.get(target).get(0).peekFirst().equals(target) ) {
			//merge
			Path p1 = nodePath.get(source).get(0);
			Path p2 = nodePath.get(target).get(0);
			if (p1.equals(p2)) {
				//add
				Path p = new Path(source, target);
				nodePath.get(source).add(p);
				nodePath.get(target).add(p);
				return;
			}
			for (String node : p2) {
				nodePath.get(node).remove(p2);
				p1.addLast(node);
				nodePath.get(node).add(p1);
			}
		}
		else if (    nodePath.get(source).size() == 1 
				  && nodePath.get(source).get(0).peekLast().equals(source)) {
			//extend
			nodePath.get(source).get(0).addLast(target);
			nodePath.get(target).add(nodePath.get(source).get(0));
		}
		else if (    nodePath.get(target).size() == 1
				  && nodePath.get(target).get(0).peekFirst().equals(target)) {
			//extend
			nodePath.get(target).get(0).addFirst(source);
			nodePath.get(source).add(nodePath.get(target).get(0));
		}
		else {
			//add
			Path p = new Path(source, target);
			nodePath.get(source).add(p);
			nodePath.get(target).add(p);			
		}
	}
	
	private void checkBlocks(String node) {
		ArrayList<Path> paths = nodePath.get(node);
		for (int i=0; i<paths.size()-1; i++) {
			Path p1 = paths.get(i);
			Block b = findBlock(p1);
			for (int j=i+1; j<paths.size(); j++) {
				Path p2 = paths.get(j);
				if (p1.equals(p2))
					continue;
				if (p1.peekFirst().equals(p2.peekFirst()) && p1.peekLast().equals(p2.peekLast())) {
					if (b == null) {
						b = new Block(p1.peekFirst(), p1.peekLast());
						addPathToBlock(b,p1);
						blocks.add(b);
					}
					addPathToBlock(b,p2);
				}					
			}
		}
	}
	private Block findBlock(Path path) {
		for (Block block : blocks) {
			if (block.getStart().equals(path.peekFirst()) && block.getEnd().equals(path.peekLast())) {
				addPathToBlock(block,path);
				return block;
			}
		}
		return null;
	}
	private void addPathToBlock(Block block, Path path) {
		block.add(path);
		for (String node : path) {
			if (!nodeBlock.containsKey(node))
				nodeBlock.put(node, new ArrayList<Block>());
			if (!nodeBlock.get(node).contains(block))
				nodeBlock.get(node).add(block);
		}
	}
	
	private void analyse(XTrace trace) {
		ArrayList<Block> currBlocks = new ArrayList<Block>();
		numBlockAsAWhole = 0;
		numBlocksInPieces = 0;
		maxSimulBlock = 0;
		for (XEvent event : trace) {
			String activity = XLogFunctions.getName(event).toUpperCase();
			if (!event.getAttributes().containsKey("id"))
				continue;
			String id = event.getAttributes().get("id").toString();
			if (    activity.equals("CREATE_ACTIVITY")
			     || activity.equals("CREATE_START_EVENT")
			     || activity.equals("CREATE_END_EVENT")
			     || activity.equals("CREATE_XOR")
			     || activity.equals("CREATE_AND")) {
				if (!nodeBlock.containsKey(id))
					continue;
				for (Block block : nodeBlock.get(id)) {
					if (!currBlocks.contains(block))
						currBlocks.add(block);
					for (Block currBlock : currBlocks)
						if (!currBlock.equals(block))
							currBlock.setNoWhole();
					if (currBlocks.size() > maxSimulBlock)
						maxSimulBlock = currBlocks.size();
					if (block.create()) {
						currBlocks.remove(block);
						if (block.isNoWhole())
							numBlocksInPieces++;
						else
							numBlockAsAWhole++;
					}
				}
			}
		}
	}
	
	public int getNumBlockAsAWhole() {
		return numBlockAsAWhole;
	}
	public int getNumBlocksInPieces() {
		return numBlocksInPieces;
	}
	public int getMaxSimulBlock() {
		return maxSimulBlock;
	}
}