package org.processmining.plugins.ppm.statistics;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileNameExtensionFilter;

import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class CalculateCheetahStatistics {
	private static final String ACT_SIZE_HEIGHT = "activity_size_height";
	private static final String ACT_SIZE_WIDTH = "activity_size_width";
	private static final String DISP_HEIGHT = "display_resolution_height";
	private static final String DISP_WIDTH = "display_resolution_width";
	private static final String DURATION = "durationInSeconds";
	private static final String EXP_WF_DUR= "experimental workflow activity duration";
	
	private static final String MODEL_ACT = "How many activities did all these models have on average?";
	private static final String YEARS_BPMN = "How many months ago did you start using BPMN? (The first version of BPMN stems from May 2004, i.e. 60 months until May 2009)";
	private static final String MODELS_ANALYSED = "How many process models have you analyzed or read within the last 12 months? (A year has about 250 work days. In case you read one model per day, this would sum up to 250 models per year)";
	private static final String MODELS_CREATED = "How many process models have you created or edited within the last 12 months?";
	private static final String FORMAL_TRAINING = "How many work days of formal training on process modeling have you received within the last 12 months? (This includes e.g. university lectures, certification courses, training courses. 15 weeks of a 90 minutes university lecture is roughly 3 work days)_duration";
	private static final String SELF_EDUCATION = "How many work days of self education have you made within the last 12 months? (This includes e.g. learning-by-doing, learning-on-the-fly, self-study of textbooks or specifications) _duration";
	private static final String YEARS_MODEL = "How many years ago did you start process modeling?";
	private static final String MENTAL_EFFORT1 = "How would you assess the mental effort for completing the first modeling task?";
	private static final String MENTAL_EFFORT2= "How would you assess the mental effort for completing the second modeling task?";
	private static final String KNOW_PREFLIGHT = "I am familiar with Pre-Flight Processes.";
	private static final String KNOW_MORTGAGE = "I am very familiar with mortgage processes.";
	private static final String EXPERT = "I consider myself being a process modeling expert.";
	private static final String RELIEF_SATISFACTION = "I feel relief rather than satisfaction after completing a task that required a lot of mental effort.";
	private static final String COMPETENT_USE_BPMN = "I feel very competent in using the BPMN for process modeling.";
	private static final String COMPETENT_GET_BPMN = "I feel very confident in understanding process models created with the BPMN.";
	private static final String LIKE_DELIBERATING = "I find satisfaction in deliberating hard and for long hours.";
	private static final String LIKE_LITTLE_THOUGHT = "I like tasks that require little thought once I&apos;ve learned them.";
	private static final String LIKE_LOT_THINKING = "I like to have the responsibility of handling a situation that requires a lot of thinking.";
	private static final String THINK_LESS = "I only think as hard as I have to.";
	private static final String LIKE_PUZZLES = "I prefer my Iife to be filled with puzzles that I must solve.";
	private static final String LIKE_SMALL_PROJECTS = "I prefer to think about small, daily projects to long-term ones.";
	private static final String LIKE_NOVELTY = "I really enjoy a task that involves coming up with new solutions to problems.";
	private static final String AVOID_THINKING = "I try to anticipate and avoid situations where there is likely chance I will have to think in depth about something.";
	private static final String USUALLY_DELIBERATE = "I usually end up deliberating about issues even when they do not affect me personally.";
	private static final String LIKE_INTELLECTUAL = "I would prefer a task that is intellectual, difficult, and important to one that is somewhat important but does not require much thought.";
	private static final String LIKE_COMPLEXITY = "I would prefer complex to simple problems.";
	private static final String PREFER_LITTLE_THOUGHT = "I would rather do something that requires little thought than something that is sure to challenge my thinking abilities.";
	private static final String HOW_IS_UNIMPORTANT = "It&apos;s enough for me that something gets the job done; I don&apos;t care how or why it works.";
	private static final String NOT_LIKE_NOVERLTY = "Learning new ways to think doesn&apos;t excite me very much.";
	private static final String FAMILIAR_BPMN = "Overall, I am very familiar with the BPMN.";
	private static final String OTHER_TECHNIQUES = "Please specify further modeling techniques you are familiar with (e.g. BPMN, Petri nets). If those techniques are closely associated to tools (e.g., Mavim, BizzDesigner), please specify those tools.";
	private static final String MOST_FAMILIAR_TECHNIQUE = "Please specify the modeling technique you are most familiar with (e.g. BPMN, Petri nets). If that technique is closely associated to a modeling tool (e.g., Mavim, BizzDesigner), please specify that tool.";
	private static final String THOUGHT_TO_TOP = "The idea of relying on thought to make my way to the top appeals to me.";
	private static final String LIKE_ABSTRACTION = "The notion of thinking abstractly is appealing to me.";
	private static final String NOT_LIKE_THINKING = "Thinking is not my idea of fun.";
	private static final String CURRENT_PROFESSION = "What is your current profession (e.g., student, consultant)?";
	private static final String WF_CONFIG_ID = "workflow_configuration_id";
	
	private static NumberFormat pctformat = NumberFormat.getPercentInstance();
	private static NumberFormat dblformat = NumberFormat.getNumberInstance();
		
	private ArrayList<String> models = new ArrayList<String>();
	private HashMap<String,String> task = new HashMap<String, String>();
	
	private CountHashMap act_size_height = new CountHashMap();
	private CountHashMap act_size_width = new CountHashMap();
	private CountHashMap disp_height = new CountHashMap();
	private CountHashMap disp_width = new CountHashMap();
	private CountHashMap duration = new CountHashMap();
	private CountHashMap exp_wf_dur = new CountHashMap();
	
	private CountHashMap model_act = new CountHashMap();
	private CountHashMap years_bpmn = new CountHashMap();
	private CountHashMap models_analysed = new CountHashMap();
	private CountHashMap models_created = new CountHashMap();
	private CountHashMap formal_training = new CountHashMap();
	private CountHashMap self_education = new CountHashMap();
	private CountHashMap years_model = new CountHashMap();

	private HashMap<String,String> mental_effort1 = new HashMap<String, String>();
	private HashMap<String,String> mental_effort2 = new HashMap<String, String>();
	private HashMap<String,String> know_preflight = new HashMap<String, String>();
	private HashMap<String,String> know_mortgage = new HashMap<String, String>();
	private HashMap<String,String> expert = new HashMap<String, String>();
	private HashMap<String,String> relief_satisfaction = new HashMap<String, String>();
	private HashMap<String,String> competent_use_bpmn = new HashMap<String, String>();
	private HashMap<String,String> competent_get_bpmn = new HashMap<String, String>();
	private HashMap<String,String> like_deliberating = new HashMap<String, String>();
	private HashMap<String,String> like_little_thought = new HashMap<String, String>();
	private HashMap<String,String> like_lot_thinking = new HashMap<String, String>();
	private HashMap<String,String> think_less = new HashMap<String, String>();
	private HashMap<String,String> like_puzzles = new HashMap<String, String>();
	private HashMap<String,String> like_small_projects = new HashMap<String, String>();
	private HashMap<String,String> like_noverlty = new HashMap<String, String>();
	private HashMap<String,String> avoid_thinking = new HashMap<String, String>();
	private HashMap<String,String> usually_deliberate = new HashMap<String, String>();
	private HashMap<String,String> like_intellectual = new HashMap<String, String>();
	private HashMap<String,String> like_complexity = new HashMap<String, String>();
	private HashMap<String,String> prefer_little_thought = new HashMap<String, String>();
	private HashMap<String,String> how_is_unimportant = new HashMap<String, String>();
	private HashMap<String,String> not_like_noverlty = new HashMap<String, String>();
	private HashMap<String,String> familiar_bpmn = new HashMap<String, String>();
	
	private HashMap<String,String> other_techniques = new HashMap<String, String>();
	private HashMap<String,String> most_familiar_technique = new HashMap<String, String>();
	private HashMap<String,String> thought_to_top = new HashMap<String, String>();
	private HashMap<String,String> like_abstraction = new HashMap<String, String>();
	private HashMap<String,String> not_like_thinking = new HashMap<String, String>();
	private HashMap<String,String> current_profession = new HashMap<String, String>();
	private HashMap<String,String> wf_config_id = new HashMap<String, String>(); 
	
	private CountHashMap num_create_activity = new CountHashMap();
	private CountHashMap num_create_and = new CountHashMap();
	private CountHashMap num_create_xor = new CountHashMap();
	private CountHashMap num_create_start_event = new CountHashMap();
	private CountHashMap num_create_end_event = new CountHashMap();
	private CountHashMap num_create_edge = new CountHashMap();
	private CountHashMap num_create_edge_bendpoint = new CountHashMap();
	private CountHashMap num_move_activity = new CountHashMap();
	private CountHashMap num_move_and = new CountHashMap();
	private CountHashMap num_move_xor = new CountHashMap();
	private CountHashMap num_move_start_event = new CountHashMap();
	private CountHashMap num_move_end_event = new CountHashMap();
	private CountHashMap num_move_edge_bendpoint = new CountHashMap();
	private CountHashMap num_move_edge_label = new CountHashMap();
	private CountHashMap num_delete_activity = new CountHashMap();
	private CountHashMap num_delete_and = new CountHashMap();
	private CountHashMap num_delete_xor = new CountHashMap();
	private CountHashMap num_delete_start_event = new CountHashMap();
	private CountHashMap num_delete_end_event = new CountHashMap();
	private CountHashMap num_delete_edge = new CountHashMap();
	private CountHashMap num_delete_edge_bendpoint = new CountHashMap();
	private CountHashMap num_reconnect_edge = new CountHashMap();
	private CountHashMap num_name_activity = new CountHashMap();
	private CountHashMap num_name_edge = new CountHashMap();
	private CountHashMap num_rename_activity = new CountHashMap();
	private CountHashMap num_rename_edge = new CountHashMap();
	
	private CountHashMap num_create_activity_before_create_end_inclgw = new CountHashMap();
	private CountHashMap num_create_activity_before_create_end_exclgw = new CountHashMap();
	private CountHashMap num_create_before_create_end_inclgw = new CountHashMap();
	private CountHashMap num_create_before_create_end_exclgw = new CountHashMap();
	private CountHashMap num_move_activity_before_create_end_inclgw = new CountHashMap();
	private CountHashMap num_move_activity_before_create_end_exclgw = new CountHashMap();
	private CountHashMap num_move_activity_before_last_update_end_inclgw = new CountHashMap();
	private CountHashMap num_move_activity_before_last_update_end_exclgw = new CountHashMap();
	
	private CountHashMap num_element_with_moves = new CountHashMap();
	private DoubleHashMap avg_num_move_move_element = new DoubleHashMap();
	private CountHashMap max_num_move_move_element = new CountHashMap();
	private DoubleHashMap avg_num_update_one_element = new DoubleHashMap();
	private CountHashMap max_num_update_one_element = new CountHashMap();
	private DoubleHashMap avg_num_consecutive_create = new DoubleHashMap();
	private CountHashMap max_num_consecutive_create = new CountHashMap();
	private DoubleHashMap avg_num_consecutive_move = new DoubleHashMap();
	private CountHashMap max_num_consecutive_move = new CountHashMap();
	private DoubleHashMap avg_num_consecutive_delete = new DoubleHashMap();
	private CountHashMap max_num_consecutive_delete = new CountHashMap();
	
	private DoubleHashMap tot_time = new DoubleHashMap();
	private DoubleHashMap tot_create_time = new DoubleHashMap();
	private DoubleHashMap avg_gap = new DoubleHashMap();
	private DoubleHashMap max_gap = new DoubleHashMap();
	private CountHashMap num_gap5s = new CountHashMap();
	private CountHashMap num_gap10s = new CountHashMap();
	private CountHashMap num_gap15s = new CountHashMap();
	private CountHashMap num_gap20s = new CountHashMap();
	private CountHashMap num_gap25s = new CountHashMap();
	private CountHashMap num_gap30s = new CountHashMap();
	private CountHashMap num_gap31s = new CountHashMap();

	private DoubleHashMap avg_intermediate_update_gateway = new DoubleHashMap();
	private CountHashMap max_intermediate_update_gateway = new CountHashMap();
	private DoubleHashMap avg_intermediate_create_gateway = new DoubleHashMap();
	private CountHashMap max_intermediate_create_gateway = new CountHashMap();
	private DoubleHashMap avg_intermediate_create_block = new DoubleHashMap();
	private CountHashMap max_intermediate_create_block = new CountHashMap(); 
	private DoubleHashMap avg_intermediate_update = new DoubleHashMap();
	private CountHashMap max_intermediate_update = new CountHashMap();
	private DoubleHashMap avg_intermediate_create = new DoubleHashMap();
	private CountHashMap max_intermediate_create = new CountHashMap(); 
	
	private CountHashMap num_block_as_a_whole = new CountHashMap();
	private CountHashMap num_blocks_in_pieces = new CountHashMap();
	private CountHashMap max_simul_block = new CountHashMap();
	
	@Plugin(name = "Calculate PPM Statistics",
			parameterLabels = { "PPM log" },
			returnLabels = { "File in filesystem" },
			returnTypes = { XLog.class },
			userAccessible = true,
			help = "Calculates statistics from a PPM log")
	@UITopiaVariant(affiliation = "Ghent University", 
			        author = "J. Claes", 
			        email = "jan.claes@ugent.be",
			        website = "http://www.janclaes.info")
	public XLog calculate(PluginContext context, XLog log) {
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(log.size());
		context.getProgress().setCaption("Analysing log...");
		context.getProgress().setIndeterminate(false);
		
		for (XTrace trace : log) {
			String key = XLogFunctions.getID(trace);
			
			models.add(key);
			task.put(key, getString(trace.getAttributes(), "PROCESS"));
			
			calculateDemographicStatistics(trace);
			calculateActivityStatistics(trace);
			calculateEndEventStatistics(trace);
			calculateConsecutiveStatistics(trace);
			calculateTimeStatistics(trace);
			calculateIntermediateStatistics(trace);
			calculateBlockStatistics(trace);
			context.getProgress().inc();
		}
		String filename = exportToCsv();
		//TODO: message look for file at...
		return log;
	}			
	
	private void calculateDemographicStatistics(XTrace trace) {
		String key = XLogFunctions.getID(trace);
		act_size_height.put(key, getInt(trace.getAttributes(),ACT_SIZE_HEIGHT));
		act_size_width.put(key, getInt(trace.getAttributes(),ACT_SIZE_WIDTH));
		disp_height.put(key, getInt(trace.getAttributes(),DISP_HEIGHT));
		disp_width.put(key, getInt(trace.getAttributes(),DISP_WIDTH));
		duration.put(key, getInt(trace.getAttributes(),DURATION));
		exp_wf_dur.put(key, getInt(trace.getAttributes(),EXP_WF_DUR));
		
		model_act.put(key, getInt(trace.getAttributes(),MODEL_ACT));
		years_bpmn.put(key, getInt(trace.getAttributes(),YEARS_BPMN));
		models_analysed.put(key, getInt(trace.getAttributes(),MODELS_ANALYSED));
		models_created.put(key, getInt(trace.getAttributes(),MODELS_CREATED));
		formal_training.put(key, getInt(trace.getAttributes(),FORMAL_TRAINING));
		self_education.put(key, getInt(trace.getAttributes(),SELF_EDUCATION));
		years_model.put(key, getInt(trace.getAttributes(),YEARS_MODEL));
	
		mental_effort1.put(key, getHighLow(trace.getAttributes(),MENTAL_EFFORT1));
		mental_effort2.put(key, getHighLow(trace.getAttributes(),MENTAL_EFFORT2));
		know_preflight.put(key, getAgreeDisagree(trace.getAttributes(),KNOW_PREFLIGHT));
		know_mortgage.put(key, getAgreeDisagree(trace.getAttributes(),KNOW_MORTGAGE));
		expert.put(key, getAgreeDisagree(trace.getAttributes(),EXPERT));
		relief_satisfaction.put(key, getAgreeDisagree(trace.getAttributes(),RELIEF_SATISFACTION));
		competent_use_bpmn.put(key, getAgreeDisagree(trace.getAttributes(),COMPETENT_USE_BPMN));
		competent_get_bpmn.put(key, getAgreeDisagree(trace.getAttributes(),COMPETENT_GET_BPMN));
		like_deliberating.put(key, getAgreeDisagree(trace.getAttributes(),LIKE_DELIBERATING));
		like_little_thought.put(key, getAgreeDisagree(trace.getAttributes(),LIKE_LITTLE_THOUGHT));
		like_lot_thinking.put(key, getAgreeDisagree(trace.getAttributes(),LIKE_LOT_THINKING));
		think_less.put(key, getAgreeDisagree(trace.getAttributes(),THINK_LESS));
		like_puzzles.put(key, getAgreeDisagree(trace.getAttributes(),LIKE_PUZZLES));
		like_small_projects.put(key, getAgreeDisagree(trace.getAttributes(),LIKE_SMALL_PROJECTS));
		like_noverlty.put(key, getAgreeDisagree(trace.getAttributes(),LIKE_NOVELTY));
		avoid_thinking.put(key, getAgreeDisagree(trace.getAttributes(),AVOID_THINKING));
		usually_deliberate.put(key, getAgreeDisagree(trace.getAttributes(),USUALLY_DELIBERATE));
		like_intellectual.put(key, getAgreeDisagree(trace.getAttributes(),LIKE_INTELLECTUAL));
		like_complexity.put(key, getAgreeDisagree(trace.getAttributes(),LIKE_COMPLEXITY));
		prefer_little_thought.put(key, getAgreeDisagree(trace.getAttributes(),PREFER_LITTLE_THOUGHT));
		how_is_unimportant.put(key, getAgreeDisagree(trace.getAttributes(),HOW_IS_UNIMPORTANT));
		not_like_noverlty.put(key, getAgreeDisagree(trace.getAttributes(),NOT_LIKE_NOVERLTY));
		familiar_bpmn.put(key, getAgreeDisagree(trace.getAttributes(),FAMILIAR_BPMN));
		
		other_techniques.put(key, getString(trace.getAttributes(),OTHER_TECHNIQUES));
		most_familiar_technique.put(key, getString(trace.getAttributes(),MOST_FAMILIAR_TECHNIQUE));
		thought_to_top.put(key, getAgreeDisagree(trace.getAttributes(),THOUGHT_TO_TOP));
		like_abstraction.put(key, getAgreeDisagree(trace.getAttributes(),LIKE_ABSTRACTION));
		not_like_thinking.put(key, getAgreeDisagree(trace.getAttributes(),NOT_LIKE_THINKING));
		current_profession.put(key, getString(trace.getAttributes(),CURRENT_PROFESSION));
		wf_config_id.put(key, getString(trace.getAttributes(),WF_CONFIG_ID));
	}
	private int getInt(XAttributeMap map, String key) {
		if (!map.containsKey(key))
			return -1;
		int result = -1;
		try {
			result = Integer.parseInt(((XAttributeLiteral)map.get(key)).getValue());
		}
		catch (ClassCastException ex) {
			result = -1;
		}
		return result;
	}
	private boolean getBoolean(XAttributeMap map, String key) {
		if (!map.containsKey(key))
			return false;
		return ((XAttributeLiteral)map.get(key)).getValue().equals("true");
	}
	private String getString(XAttributeMap map, String key) {
		if (!map.containsKey(key))
			return "";
		return ((XAttributeLiteral)map.get(key)).getValue();
	}
	private String getHighLow(XAttributeMap map, String key) {
		String res = "";
		if (map.containsKey(key + "_High") && getBoolean(map, key + "_High"))
			res += "High";
		if (map.containsKey(key + "_Low") && getBoolean(map, key + "_Low"))
			res += "Low";
		if (map.containsKey(key + "_Medium") && getBoolean(map, key + "_Medium"))
			res += "Medium";
		if (map.containsKey(key + "_Rather High") && getBoolean(map, key + "_Rather High"))
			res += "Rather High";
		if (map.containsKey(key + "_Rather Low") && getBoolean(map, key + "_Rather Low"))
			res += "Rather Low";
		if (map.containsKey(key + "_Very High") && getBoolean(map, key + "_Very High"))
			res += "Very High";
		if (map.containsKey(key + "_Very Low") && getBoolean(map, key + "_Very Low"))
			res += "Very Low";
		return res;
	}
	private String getAgreeDisagree(XAttributeMap map, String key) {
		String res = "";
		if (map.containsKey(key + "_agree") && getBoolean(map, key + "_agree"))
			res += "agree";
		if (map.containsKey(key + "_disagree") && getBoolean(map, key + "_disagree"))
			res += "disagree";
		if (map.containsKey(key + "_neutral") && getBoolean(map, key + "_neutral"))
			res += "neutral";
		if (map.containsKey(key + "_somewhat agree") && getBoolean(map, key + "_somewhat agree"))
			res += "somewhat agree";
		if (map.containsKey(key + "_somewhat disagree") && getBoolean(map, key + "_somewhat disagree"))
			res += "somewhat disagree";
		if (map.containsKey(key + "_strongly agree") && getBoolean(map, key + "_strongly agree"))
			res += "strongly agree";
		if (map.containsKey(key + "_strongly disagree") && getBoolean(map, key + "_strongly disagree"))
			res += "strongly disagree";
		return res;
	}
/*	
	What is your current profession (e.g., student, consultant)?" value="Product Manager"/>
	What is your current profession (e.g., student, consultant)?_duration" value="10281"/>
	workflow_configuration_id" value="1682"/>
*/
			
	private void calculateActivityStatistics(XTrace trace) {
		String key = XLogFunctions.getID(trace);
		for (XEvent event : trace) {
			String activity = XLogFunctions.getName(event);
			if (getNumMap(activity) != null) 
				getNumMap(activity).incr(key);
		}
	}
	
	private void calculateEndEventStatistics(XTrace trace) {
		String key = XLogFunctions.getID(trace);
		boolean first_create_end_reached = false;
		boolean last_update_end_reached = false;
		String last_update_end = searchLastUpdateEnd(trace);
		for (XEvent event : trace) {
			String id = getID(event);
			String activity = XLogFunctions.getName(event);
			
			if (isCreate(activity) && isEnd(activity))
				first_create_end_reached = true;
			if (id.equals(last_update_end))
				last_update_end_reached = true;
			
			if (!first_create_end_reached && isCreate(activity) && (isActivity(activity) || isGateway(activity))) {
				if (!isGateway(activity))
					num_create_activity_before_create_end_exclgw.incr(key);
				num_create_activity_before_create_end_inclgw.incr(key);
			}
			if (!first_create_end_reached && isCreate(activity)) {
				if (!isGateway(activity))
					num_create_before_create_end_exclgw.incr(key);
				num_create_before_create_end_inclgw.incr(key);
			}
			if (!first_create_end_reached && isMove(activity) && (isActivity(activity) || isGateway(activity))) {
				if (!isGateway(activity))
					num_move_activity_before_create_end_exclgw.incr(key);
				num_move_activity_before_create_end_inclgw.incr(key);
			}
			if (!last_update_end_reached && isMove(activity) && (isActivity(activity) || isGateway(activity))) {
				if (!isGateway(activity))
					num_move_activity_before_last_update_end_exclgw.incr(key);
				num_move_activity_before_last_update_end_inclgw.incr(key);
			}
		}
	}
	private String searchLastUpdateEnd(XTrace trace) {
		String key = "";
		for (XEvent event : trace)
		{
			String activity = XLogFunctions.getName(event);
			String id = getID(event);
			if (isEnd(activity))
				key = id;
		}
		return key;
	}
	
	private void calculateConsecutiveStatistics(XTrace trace) {		
		String key = XLogFunctions.getID(trace);
		CountHashMap updates = new CountHashMap();
		CountHashMap moves = new CountHashMap();
		String lastAction = "";
		
		int curr_consecutive_create = 0;
		int max_consecutive_create = 0;
		int total_consecutive_create = 0;
		int num_consecutive_create = 0;
		
		int curr_consecutive_move = 0;
		int max_consecutive_move = 0;
		int total_consecutive_move = 0;
		int num_consecutive_move = 0;
		
		int curr_consecutive_delete = 0;
		int max_consecutive_delete = 0;
		int total_consecutive_delete = 0;
		int num_consecutive_delete = 0;

		for (XEvent event : trace) {
			String id = getID(event);
			String activity = XLogFunctions.getName(event);
			
			updates.incr(id);
			if (isMove(activity))
				moves.incr(id);
			
			if (lastAction.equals("C")) {
				if (isCreate(activity)) {
					curr_consecutive_create++;
				}
				else if (curr_consecutive_create > 0) {
					if (curr_consecutive_create > max_consecutive_create)
						max_consecutive_create = curr_consecutive_create;
					total_consecutive_create += curr_consecutive_create;
					num_consecutive_create++;
					curr_consecutive_create = 0;
					lastAction = "C";
				}
			}
			else if (lastAction.equals("M")) {
				if (isMove(activity)) {
					curr_consecutive_move++;
				}
				else if (curr_consecutive_move > 0) {
					if (curr_consecutive_move > max_consecutive_move)
						max_consecutive_move = curr_consecutive_move;
					total_consecutive_move += curr_consecutive_move;
					num_consecutive_move++;
					curr_consecutive_move = 0;
					lastAction = "M";
				}
			}
			if (lastAction.equals("D")) {
				if (isDelete(activity)) {
					curr_consecutive_delete++;
				}
				else if (curr_consecutive_delete > 0) {
					if (curr_consecutive_delete > max_consecutive_delete)
						max_consecutive_delete = curr_consecutive_delete;
					total_consecutive_delete += curr_consecutive_delete;
					num_consecutive_delete++;
					curr_consecutive_delete = 0;
					lastAction = "D";
				}
			}
		}

		num_element_with_moves.put(key, moves.size());

		double avg = 0.0;
		int max = 0;
		Object[] keys = moves.keySet().toArray();
		for (int i=0; i<keys.length; i++) {
			avg += moves.get(keys[i]);
			if (moves.get(keys[i]) > max)
				max = moves.get(keys[i]);
		}
		avg_num_move_move_element.put(key, avg * 1.0 / keys.length);
		max_num_move_move_element.put(key, max);
		
		avg = 0.0;
		max = 0;
		keys = updates.keySet().toArray();
		for (int i=0; i<keys.length; i++) {
			avg += updates.get(keys[i]);
			if (updates.get(keys[i]) > max)
				max = updates.get(keys[i]);
		}
		avg_num_update_one_element.put(key, avg * 1.0 / keys.length);
		max_num_update_one_element.put(key, max);
		
		avg_num_consecutive_create.put(key, num_consecutive_create == 0 ? -1 : (total_consecutive_create * 1.0 / num_consecutive_create));
		max_num_consecutive_create.put(key, max_consecutive_create);
		avg_num_consecutive_move.put(key, num_consecutive_move == 0 ? -1 : (total_consecutive_move * 1.0 / num_consecutive_move));
		max_num_consecutive_move.put(key, max_consecutive_move);
		avg_num_consecutive_delete.put(key, num_consecutive_delete == 0 ? -1 : (total_consecutive_delete * 1.0 / num_consecutive_delete));
		max_num_consecutive_delete.put(key, max_consecutive_delete);
	}
	
	
	private void calculateTimeStatistics(XTrace trace) {
		String key = XLogFunctions.getID(trace);
		long start = XLogFunctions.getTime(trace.get(0)).getTime();
		long end = XLogFunctions.getTime(trace.get(trace.size()-1)).getTime();
		long create_end = start;
		long curr = 0, prev = 0;
		double avg = 0, max = 0;

		tot_time.put(key, (end - start)/1000.0);

		boolean first = true;
		for (XEvent event : trace) {
			String activity = XLogFunctions.getName(event);
			prev = curr;
			curr = XLogFunctions.getTime(event).getTime();

			if (isCreate(activity))
				create_end = curr;
			
			if (!first) {
				double gap = (curr-prev)/1000.0;
				if (gap > max)
					max = gap;
				avg += gap;
				
				if (gap < 5)
					num_gap5s.incr(key);
				else if (gap < 10)
					num_gap10s.incr(key);
				else if (gap < 15)
					num_gap15s.incr(key);
				else if (gap < 20)
					num_gap20s.incr(key);
				else if (gap < 25)
					num_gap25s.incr(key);
				else if (gap < 30)
					num_gap30s.incr(key);
				else
					num_gap31s.incr(key);
			}
			first = false;
		}
		tot_create_time.put(key, (create_end - start)/1000.0);
		avg_gap.put(key, avg / (trace.size() - 1));
		max_gap.put(key, max);
	}
			
	private void calculateIntermediateStatistics(XTrace trace) {
		String key = XLogFunctions.getID(trace);
		int max_update_gw = 0, tot_update_gw = 0, num_update_gw = 0;
		int max_create_gw = 0, tot_create_gw = 0, num_create_gw = 0;
		int max_create_block = 0, tot_create_block = 0, num_create_block = 0;
		int max_update = 0, tot_update = 0, num_update = 0;
		int max_create = 0, tot_create = 0, num_create = 0;

		FlowNetwork net = new FlowNetwork(trace);
		
		int j = 0;
		for (XEvent event : trace) {
			String id = getID(event);
			String activity = XLogFunctions.getName(event);
			
			if (isGateway(activity)) {
				CountHashMap id2_update = net.findMatchingGateways(id);
				CountHashMap id2_create = (CountHashMap)id2_update.clone();
				
				ArrayList<String> ids_create = null;
				if (id2_create.size() == 1) {
					ids_create = net.findMatchingBlock(id, id2_create.keySet().toArray()[0].toString());
					boolean inBlock = false;
					int numCreatesInBlock = 0, totalCreatesInBlock = 0;
					for (int k=0; k<trace.size(); k++) {
						String currId = getID(trace.get(k)); 
						String currActivity = XLogFunctions.getName(trace.get(k));
						if (!isCreate(currActivity))
							continue;
						if (!inBlock && !ids_create.contains(currId))
							continue;
						
						inBlock = true;
						if (!ids_create.contains(currId))
							numCreatesInBlock++;
						else {
							totalCreatesInBlock += numCreatesInBlock;
							numCreatesInBlock = 0;
						}
					}
					if (inBlock) {
						if (totalCreatesInBlock > max_create_block)
							max_create_block = totalCreatesInBlock;
						tot_create_block += totalCreatesInBlock;
						num_create_block++;
					}
				}
				
				for (int k=j+1; k<trace.size(); k++) {
					String currActivity = XLogFunctions.getName(trace.get(k));
					String currId = getID(trace.get(k));
					if (isCreate(currActivity) && id2_update.containsKey(currId)) {
						int val_update = id2_update.get(currId);
						if (val_update > max_update_gw)
							max_update_gw = val_update;
						tot_update_gw += val_update;
						num_update_gw++;
						id2_update.remove(currId);
						int val_create = id2_create.get(currId);
						if (val_create > max_create_gw)
							max_create_gw = val_create;
						tot_create_gw += val_create;
						num_create_gw++;
						id2_create.remove(currId);
					}
					Object[] keys = id2_update.keySet().toArray();
					if (keys.length == 0)
						break;
					for (int gw=0; gw<keys.length; gw++) {
						id2_update.incr((String)keys[gw]);
						if (isCreate(currActivity))
							id2_create.incr((String)keys[gw]);
					}
				}
			}
			
			int update = 0, create = 0;
			for (int k=j+1; k<trace.size(); k++) {
				String currActivity = XLogFunctions.getName(trace.get(k));
				String currId = getID(trace.get(k));
				if (id.equals(currId))
					break;
				update++;
				if (isCreate(currActivity))
					create++;
			}
			if (update > max_update)
				max_update = update;
			tot_update += update;
			num_update++;
			if (create > max_create)
				max_create = create;
			tot_create += create;
			num_create++;
			j++;
		}
		avg_intermediate_update_gateway.put(key, tot_update_gw * 1.0 / num_update_gw);
		max_intermediate_update_gateway.put(key, max_update_gw);
		avg_intermediate_create_gateway.put(key, tot_create_gw * 1.0 / num_create_gw);
		max_intermediate_create_gateway.put(key, max_create_gw);
		avg_intermediate_create_block.put(key, tot_create_block * 1.0 / num_create_block);
		max_intermediate_create_block.put(key, max_create_block);
		avg_intermediate_update.put(key, tot_update * 1.0 / num_update);
		max_intermediate_update.put(key, max_update);
		avg_intermediate_create.put(key, tot_create * 1.0 / num_create);
		max_intermediate_create.put(key, max_create);
	}
	private void calculateBlockStatistics(XTrace trace) {
		String key = XLogFunctions.getID(trace);
		if (key.equals("71"))
			key = "71";
		BlockAnalyser ba = new BlockAnalyser(trace);
		num_block_as_a_whole.put(key, ba.getNumBlockAsAWhole());
		num_blocks_in_pieces.put(key, ba.getNumBlocksInPieces());
		max_simul_block.put(key, ba.getMaxSimulBlock());
	}
	
	private String exportToCsv() {
		JFileChooser fc = new JFileChooser();
		fc.setFileFilter(new FileNameExtensionFilter("*.csv files", "csv"));
		if (fc.showSaveDialog(null) == JFileChooser.APPROVE_OPTION) {
			try {
				//TODO ask overwrite if file exists
				File file = fc.getSelectedFile();
				FileWriter fw = new FileWriter(file);
				fw.write("CaseNum;Task;");
				
				writeNumEventsHeading(fw);
				writeNumCreateEventsHeading(fw);
				writeNumMoveEventsHeading(fw);
				writeNumDeleteEventsHeading(fw);
				writeNumEndPositionHeading(fw);
				writeNumConsecutiveEventsHeading(fw);
				writeTimeHeading(fw);
				writeIntermediateHeading(fw);
				writeBlockHeading(fw);
				
				writePercEventsHeading(fw);
				writePercCreateEventsHeading(fw);
				writePercMoveEventsHeading(fw);
				writePercDeleteEventsHeading(fw);
				writePercEndPositionHeading(fw);
				writePercConsecutiveEventsHeading(fw);
				writePercTimeHeading(fw);
				writePercIntermediateHeading(fw);
				writePercBlockHeading(fw);
				
				writeDemographicHeading(fw);
				
				fw.append("\n");
				
				for (String key : models) {
					fw.append(key + ";");
					fw.append(task.get(key) + ";");
					
					writeNumEvents(fw, key);
					writeNumCreateEvents(fw, key);
					writeNumMoveEvents(fw, key);
					writeNumDeleteEvents(fw, key);
					writeNumEndPosition(fw, key);
					writeNumConsecutiveEvents(fw, key);
					writeTime(fw, key);
					writeIntermediate(fw, key);
					writeBlock(fw, key);
					
					writePercEvents(fw, key);
					writePercCreateEvents(fw, key);
					writePercMoveEvents(fw, key);
					writePercDeleteEvents(fw, key);
					writePercEndPosition(fw, key);
					writePercConsecutiveEvents(fw, key);
					writePercTime(fw, key);
					writePercIntermediate(fw, key);
					writePercBlock(fw, key);
					
					writeDemographic(fw, key);
					
					fw.append("\n");
				}
				fw.close();
				return file.getPath();
			}
			catch (Exception ex) {
				JOptionPane.showMessageDialog(null, "Error trying to save results to file", "Error", JOptionPane.ERROR_MESSAGE);
			}
		}
		return "ERROR";
	}
	
	private void writeNumEventsHeading(FileWriter fw) throws IOException {
		fw.append("NumEvent;");
		fw.append("NumCreate;");
		fw.append("NumMove;");
		fw.append("NumDelete;");
	}
	private void writeNumEvents(FileWriter fw, String key) throws IOException {
		fw.append(numEvent(key) + ";");
		fw.append(numCreate(key) + ";");
		fw.append(numMove(key) + ";");
		fw.append(numDelete(key) + ";");
	}
	private void writePercEventsHeading(FileWriter fw) throws IOException {
		fw.append("PercCreate;");
		fw.append("PercMove;");
		fw.append("PercDelete;");
	}
	private void writePercEvents(FileWriter fw, String key) throws IOException {
		fw.append(irt(numCreate(key),numEvent(key)) + ";");
		fw.append(irt(numMove(key),numEvent(key)) + ";");
		fw.append(irt(numDelete(key),numEvent(key)) + ";");
	}

	private void writeNumCreateEventsHeading(FileWriter fw) throws IOException {
		fw.append("NumCreateActivity;");
		fw.append("NumCreateAnd;");
		fw.append("NumCreateXor;");
		fw.append("NumCreateEnd;");
		fw.append("NumCreateStart;");
		fw.append("NumCreateGateway;");
		fw.append("NumCreateEdge;");
		fw.append("NumCreate;");
	}
	private void writeNumCreateEvents(FileWriter fw, String key) throws IOException {
		fw.append(num_create_activity.get(key) + ";");
		fw.append(num_create_and.get(key) + ";");
		fw.append(num_create_xor.get(key) + ";");
		fw.append(num_create_end_event.get(key) + ";");
		fw.append(num_create_start_event.get(key) + ";");
		fw.append((num_create_and.get(key)+ num_create_xor.get(key)) + ";");
		fw.append(num_create_edge.get(key) + ";");
		fw.append(numCreate(key) + ";");
	}
	private void writePercCreateEventsHeading(FileWriter fw) throws IOException {
		fw.append("PercCreateActivity;");
		fw.append("PercCreateAnd;");
		fw.append("PercCreateXor;");
		fw.append("PercCreateEnd;");
		fw.append("PercCreateStart;");
		fw.append("PercCreateGateway;");
		fw.append("PercCreateEdge;");
		fw.append("PercCreate;");
	}
	private void writePercCreateEvents(FileWriter fw, String key) throws IOException {
		fw.append(irt(num_create_activity.get(key),numCreate(key)) + ";");
		fw.append(irt(num_create_and.get(key),numCreate(key)) + ";");
		fw.append(irt(num_create_xor.get(key),numCreate(key)) + ";");
		fw.append(irt(num_create_end_event.get(key),numCreate(key)) + ";");
		fw.append(irt(num_create_start_event.get(key),numCreate(key)) + ";");
		fw.append(irt(num_create_and.get(key)+ num_create_xor.get(key),numCreate(key)) + ";");
		fw.append(irt(num_create_edge.get(key),numCreate(key)) + ";");
		fw.append(irt(numCreate(key),numEvent(key)) + ";");
	}
	
	private void writeNumMoveEventsHeading(FileWriter fw) throws IOException {
		fw.append("NumMoveActivity;");
		fw.append("NumMoveAnd;");
		fw.append("NumMoveXor;");
		fw.append("NumMoveEnd;");
		fw.append("NumMoveStart;");
		fw.append("NumMoveGateway;");
		fw.append("NumMoveEdge;");
		fw.append("NumMove;");
	}
	private void writeNumMoveEvents(FileWriter fw, String key) throws IOException {
		fw.append(num_move_activity.get(key) + ";");
		fw.append(num_move_and.get(key) + ";");
		fw.append(num_move_xor.get(key) + ";");
		fw.append(num_move_end_event.get(key) + ";");
		fw.append(num_move_start_event.get(key) + ";");
		fw.append((num_move_and.get(key)+ num_move_xor.get(key)) + ";");
		fw.append((num_create_edge_bendpoint.get(key) + num_move_edge_bendpoint.get(key) + num_delete_edge_bendpoint.get(key) + num_move_edge_label.get(key)) + ";");
		fw.append(numMove(key) + ";");
	}
	private void writePercMoveEventsHeading(FileWriter fw) throws IOException {
		fw.append("PercMoveActivity;");
		fw.append("PercMoveAnd;");
		fw.append("PercMoveXor;");
		fw.append("PercMoveEnd;");
		fw.append("PercMoveStart;");
		fw.append("PercMoveGateway;");
		fw.append("PercMoveEdge;");
		fw.append("PercMove;");
	}
	private void writePercMoveEvents(FileWriter fw, String key) throws IOException {
		fw.append(irt(num_move_activity.get(key),numMove(key)) + ";");
		fw.append(irt(num_move_and.get(key),numMove(key)) + ";");
		fw.append(irt(num_move_xor.get(key),numMove(key)) + ";");
		fw.append(irt(num_move_end_event.get(key),numMove(key)) + ";");
		fw.append(irt(num_move_start_event.get(key),numMove(key)) + ";");
		fw.append(irt(num_move_and.get(key)+ num_move_xor.get(key),numMove(key)) + ";");
		fw.append(irt(num_create_edge_bendpoint.get(key) + num_move_edge_bendpoint.get(key) + num_delete_edge_bendpoint.get(key) + num_move_edge_label.get(key),numMove(key)) + ";");
		fw.append(irt(numMove(key),numEvent(key)) + ";");
	}
	
	private void writeNumDeleteEventsHeading(FileWriter fw) throws IOException {
		fw.append("NumDeleteActivity;");
		fw.append("NumDeleteAnd;");
		fw.append("NumDeleteXor;");
		fw.append("NumDeleteEnd;");
		fw.append("NumDeleteStart;");
		fw.append("NumDeleteGateway;");
		fw.append("NumDeleteEdge;");
		fw.append("NumDelete;");
	}
	private void writeNumDeleteEvents(FileWriter fw, String key) throws IOException {
		fw.append(num_delete_activity.get(key) + ";");
		fw.append(num_delete_and.get(key) + ";");
		fw.append(num_delete_xor.get(key) + ";");
		fw.append(num_delete_end_event.get(key) + ";");
		fw.append(num_delete_start_event.get(key) + ";");
		fw.append((num_delete_and.get(key)+ num_delete_xor.get(key)) + ";");
		fw.append(num_delete_edge.get(key) + ";");
		fw.append(numDelete(key) + ";");
	}
	private void writePercDeleteEventsHeading(FileWriter fw) throws IOException {
		fw.append("PercDeleteActivity;");
		fw.append("PercDeleteAnd;");
		fw.append("PercDeleteXor;");
		fw.append("PercDeleteEnd;");
		fw.append("PercDeleteStart;");
		fw.append("PercDeleteGateway;");
		fw.append("PercDeleteEdge;");
		fw.append("PercDelete;");
	}
	private void writePercDeleteEvents(FileWriter fw, String key) throws IOException {
		fw.append(irt(num_delete_activity.get(key),numDelete(key)) + ";");
		fw.append(irt(num_delete_and.get(key),numDelete(key)) + ";");
		fw.append(irt(num_delete_xor.get(key),numDelete(key)) + ";");
		fw.append(irt(num_delete_end_event.get(key),numDelete(key)) + ";");
		fw.append(irt(num_delete_start_event.get(key),numDelete(key)) + ";");
		fw.append(irt(num_delete_and.get(key)+ num_delete_xor.get(key),numDelete(key)) + ";");
		fw.append(irt(num_delete_edge.get(key),numDelete(key)) + ";");
		fw.append(irt(numDelete(key),numEvent(key)) + ";");
	}
	
	private void writeNumEndPositionHeading(FileWriter fw) throws IOException {
		fw.append("NumCreateActivityBeforeCreateEndInclGW;NumCreateActivityBeforeCreateEndExclGW;");
		fw.append("NumCreateBeforeCreateEndInclGW;NumCreateBeforeCreateEndExclGW;");
		fw.append("NumMoveActivityBeforeCreateEndInclGW;NumMoveActivityBeforeCreateEndExclGW;");
		fw.append("NumMoveActivityBeforeLastUpdateEndInclGW;NumMoveActivityBeforeLastUpdateEndExclGW;");
	}
	private void writeNumEndPosition(FileWriter fw, String key) throws IOException {
		fw.append(num_create_activity_before_create_end_inclgw.get(key) + ";" + num_create_activity_before_create_end_exclgw.get(key) + ";");
		fw.append(num_create_before_create_end_inclgw.get(key) + ";" + num_create_before_create_end_exclgw.get(key) + ";");
		fw.append(num_move_activity_before_create_end_inclgw.get(key) + ";" + num_move_activity_before_create_end_exclgw.get(key) + ";");
		fw.append(num_move_activity_before_last_update_end_inclgw.get(key) + ";" + num_move_activity_before_last_update_end_exclgw.get(key) + ";");
	}
	private void writePercEndPositionHeading(FileWriter fw) throws IOException {
		fw.append("PercCreateActivityBeforeCreateEndInclGW;PercCreateActivityBeforeCreateEndExclGW;");
		fw.append("PercCreateBeforeCreateEndInclGW;PercCreateBeforeCreateEndExclGW;");
		fw.append("PercMoveActivityBeforeCreateEndInclGW;PercMoveActivityBeforeCreateEndExclGW;");
		fw.append("PercMoveActivityBeforeLastUpdateEndInclGW;PercMoveActivityBeforeLastUpdateEndExclGW;");
	}
	private void writePercEndPosition(FileWriter fw, String key) throws IOException {
		fw.append(irt(num_create_activity_before_create_end_inclgw.get(key), numCreate(key)) + ";" + irt(num_create_activity_before_create_end_exclgw.get(key), numCreate(key)) + ";");
		fw.append(irt(num_create_before_create_end_inclgw.get(key), numCreate(key)) + ";" + irt(num_create_before_create_end_exclgw.get(key), numCreate(key)) + ";");
		fw.append(irt(num_move_activity_before_create_end_inclgw.get(key), numMove(key)) + ";" + irt(num_move_activity_before_create_end_exclgw.get(key), numMove(key)) + ";");
		fw.append(irt(num_move_activity_before_last_update_end_inclgw.get(key), numMove(key)) + ";" + irt(num_move_activity_before_last_update_end_exclgw.get(key), numMove(key)) + ";");
	}

	private void writeNumConsecutiveEventsHeading(FileWriter fw) throws IOException {
		fw.append("NumElementsWithMoves;");
		fw.append("AvgMoveOnMovedElements;MaxMoveOnMoveElements;");
		fw.append("AvgUpdateOneElemet;MaxUpdateOneElement;");
		fw.append("AvgConsecutiveCreate;MaxConsecutiveCreate;");
		fw.append("AvgConsecutiveMove;MaxConsecutiveMove;");
		fw.append("AvgConsecutiveDelete;MaxConsecutiveDelete;");
	}
	private void writeNumConsecutiveEvents(FileWriter fw, String key) throws IOException {
		fw.append(num_element_with_moves.get(key) + ";");
		fw.append(round(avg_num_move_move_element.get(key)) + ";" + max_num_move_move_element.get(key) + ";");
		fw.append(round(avg_num_update_one_element.get(key)) + ";" + max_num_update_one_element.get(key) + ";");
		fw.append(round(avg_num_consecutive_create.get(key)) + ";" + max_num_consecutive_create.get(key) + ";");
		fw.append(round(avg_num_consecutive_move.get(key)) + ";" + max_num_consecutive_move.get(key) + ";");
		fw.append(round(avg_num_consecutive_delete.get(key)) + ";" + max_num_consecutive_delete.get(key) + ";");
	}
	private void writePercConsecutiveEventsHeading(FileWriter fw) throws IOException {
		fw.append("PercNumElementsWithMoves;");
		fw.append("PercAvgMoveOnMovedElements;PercMaxMoveOnMoveElements;");
		fw.append("PercAvgUpdateOneElemet;PercMaxUpdateOneElement;" );
		fw.append("PercAvgConsecutiveCreateIRTCreate;PercAvgConsecutiveCreateIRTTotal;PercMaxConsecutiveCreateIRTCreate;PercMaxConsecutiveCreateIRTTotal;");
		fw.append("PercAvgConsecutiveMoveIRTMove;PercAvgConsecutiveMoveIRTTotal;PercMaxConsecutiveMoveIRTMove;PercMaxConsecutiveMoveIRTTotal;");
		fw.append("PercAvgConsecutiveDeleteIRTDelete;PercAvgConsecutiveDeleteIRTTotal;PercMaxConsecutiveDeleteIRTDelete;PercMaxConsecutiveDeleteIRTTotal;");
	}
	private void writePercConsecutiveEvents(FileWriter fw, String key) throws IOException {
		fw.append(irt(num_element_with_moves.get(key),numCreate(key)) + ";");
		fw.append(irt(avg_num_move_move_element.get(key), numMove(key)) + ";" + irt(max_num_move_move_element.get(key),numMove(key)) + ";");
		fw.append(irt(avg_num_update_one_element.get(key),numEvent(key)) + ";" + irt(max_num_update_one_element.get(key),numEvent(key)) + ";");
		fw.append(irt(avg_num_consecutive_create.get(key),numCreate(key)) + ";" + irt(avg_num_consecutive_create.get(key),numEvent(key)) + ";" + irt(max_num_consecutive_create.get(key),numCreate(key)) + ";" + irt(max_num_consecutive_create.get(key),numEvent(key)) + ";");
		fw.append(irt(avg_num_consecutive_move.get(key),numMove(key)) + ";" + irt(avg_num_consecutive_move.get(key),numEvent(key)) + ";" + irt(max_num_consecutive_move.get(key),numMove(key)) + ";" + irt(max_num_consecutive_move.get(key),numEvent(key)) + ";");
		fw.append(irt(avg_num_consecutive_delete.get(key),numDelete(key)) + ";" + irt(avg_num_consecutive_delete.get(key),numEvent(key)) + ";" + irt(max_num_consecutive_delete.get(key),numDelete(key)) + ";" + irt(max_num_consecutive_delete.get(key),numEvent(key)) + ";");
	}
	
	private void writeTimeHeading(FileWriter fw) throws IOException {
		fw.append("TotTime;TotCreateTime;");
		fw.append("AvgGap;MaxGap;");
		fw.append("NumGap<5s;NumGap<10s;NumGap<15s;NumGap<20s;NumGap<25s;NumGap<30s;NumGap>30s;");
	}
	private void writeTime(FileWriter fw, String key) throws IOException {
		fw.append(time(tot_time.get(key)) + ";" + time(tot_create_time.get(key)) + ";");
		fw.append(time(avg_gap.get(key)) + ";" + time(max_gap.get(key)) + ";");
		fw.append(num_gap5s.get(key) + ";" + num_gap10s.get(key) + ";" + num_gap15s.get(key) + ";" + num_gap20s.get(key) + ";" + num_gap25s.get(key) + ";" + num_gap30s.get(key) + ";" + num_gap31s.get(key) + ";");
	}
	private void writePercTimeHeading(FileWriter fw) throws IOException {
		fw.append("PercTotCreateTime;");
		fw.append("PercAvgGap;PercMaxGap;");
		fw.append("PercNumGap<5s;PercNumGap<10s;PercNumGap<15s;PercNumGap<20s;PercNumGap<25s;PercNumGap<30s;PercNumGap>30s;");
	}
	private void writePercTime(FileWriter fw, String key) throws IOException {
		fw.append(irt(tot_create_time.get(key), tot_time.get(key)) + ";");
		fw.append(irt(avg_gap.get(key), tot_time.get(key)) + ";" + irt(max_gap.get(key), tot_time.get(key)) + ";");
		fw.append(irt(num_gap5s.get(key), numEvent(key)) + ";" + irt(num_gap10s.get(key), numEvent(key)) + ";" + irt(num_gap15s.get(key), numEvent(key)) + ";" + irt(num_gap20s.get(key), numEvent(key)) + ";" + irt(num_gap25s.get(key), numEvent(key)) + ";" + irt(num_gap30s.get(key), numEvent(key)) + ";" + irt(num_gap31s.get(key), numEvent(key)) + ";");
	}
	
	private void writeIntermediateHeading(FileWriter fw) throws IOException {
		fw.append("AvgIntermediateUpdateGateway;MaxIntermediateUpdateGateway;");;
		fw.append("AvgIntermediateCreateGateway;MaxIntermediateCreateGateway;");
		fw.append("AvgIntermediateCreateBlock;MaxIntermediateCreateBlock;");
		fw.append("AvgIntermediateUpdate;MaxIntermediateUpdate;");
		fw.append("AvgIntermediateCreate;MaxIntermediateCreate;");
	}
	private void writeIntermediate(FileWriter fw, String key) throws IOException {
		fw.append(round(avg_intermediate_update_gateway.get(key)) + ";" + max_intermediate_update_gateway.get(key) + ";");
		fw.append(round(avg_intermediate_create_gateway.get(key)) + ";" + max_intermediate_create_gateway.get(key) + ";");
		fw.append(round(avg_intermediate_create_block.get(key)) + ";" + max_intermediate_create_block.get(key) + ";");
		fw.append(round(avg_intermediate_update.get(key)) + ";" + max_intermediate_update.get(key) + ";");
		fw.append(round(avg_intermediate_create.get(key)) + ";" + max_intermediate_create.get(key) + ";");
	}
	private void writePercIntermediateHeading(FileWriter fw) throws IOException {
		fw.append("PercAvgIntermediateUpdateGateway;PercMaxIntermediateUpdateGateway;");
		fw.append("PercAvgIntermediateCreateGateway;PercMaxIntermediateCreateGateway;");
		fw.append("PercAvgIntermediateCreateBlock;PercMaxIntermediateCreateBlock;");
		fw.append("PercAvgIntermediateUpdate;PercMaxIntermediateUpdate;");
		fw.append("PercAvgIntermediateCreate;PercMaxIntermediateCreate;");
	}
	private void writePercIntermediate(FileWriter fw, String key) throws IOException {
		fw.append(irt(avg_intermediate_update_gateway.get(key),numEvent(key)) + ";" + irt(max_intermediate_update_gateway.get(key),numEvent(key)) + ";");
		fw.append(irt(avg_intermediate_create_gateway.get(key),numCreate(key)) + ";" + irt(max_intermediate_create_gateway.get(key),numCreate(key)) + ";");
		fw.append(irt(avg_intermediate_create_block.get(key),numCreate(key)) + ";" + irt(max_intermediate_create_block.get(key),numCreate(key)) + ";");
		fw.append(irt(avg_intermediate_update.get(key),numEvent(key)) + ";" + irt(max_intermediate_update.get(key),numEvent(key)) + ";");
		fw.append(irt(avg_intermediate_create.get(key),numCreate(key)) + ";" + irt(max_intermediate_create.get(key),numCreate(key)) + ";");
	}

	private void writeBlockHeading(FileWriter fw) throws IOException {
		fw.append("NumBlockAsAWhole;NumBlocksInPieces;");
		fw.append("MaxSimulBlock;");
	}
	private void writeBlock(FileWriter fw, String key) throws IOException {
		fw.append(num_block_as_a_whole.get(key) + ";" + num_blocks_in_pieces.get(key) + ";");
		fw.append(max_simul_block.get(key) + ";");
	}
	private void writePercBlockHeading(FileWriter fw) throws IOException {
		fw.append("PercNumBlockAsAWhole;PercNumBlocksInPieces;");
	}
	private void writePercBlock(FileWriter fw, String key) throws IOException {
		int num = num_block_as_a_whole.get(key) + num_blocks_in_pieces.get(key);
		fw.append(irt(num_block_as_a_whole.get(key),num) + ";" + irt(num_blocks_in_pieces.get(key),num) + ";");
	}
	
	private void writeDemographicHeading(FileWriter fw) throws IOException {
		fw.append("act_size_height;");
		fw.append("act_size_width;");
		fw.append("disp_height;");
		fw.append("disp_width;");
		fw.append("duration;");
		fw.append("exp_wf_dur;");
		
		fw.append("model_act;");
		fw.append("years_bpmn;");
		fw.append("models_analysed;");
		fw.append("models_created;");
		fw.append("formal_training;");
		fw.append("self_education;");
		fw.append("years_model;");

		fw.append("mental_effort1;");
		fw.append("mental_effort2;");
		fw.append("know_preflight;");
		fw.append("know_mortgage;");
		fw.append("expert;");
		fw.append("relief_satisfaction;");
		fw.append("competent_use_bpmn;");
		fw.append("competent_get_bpmn;");
		fw.append("like_deliberating;");
		fw.append("like_little_thought;");
		fw.append("like_lot_thinking;");
		fw.append("think_less;");
		fw.append("like_puzzles;");
		fw.append("like_small_projects;");
		fw.append("like_noverlty;");
		fw.append("avoid_thinking;");
		fw.append("usually_deliberate;");
		fw.append("like_intellectual;");
		fw.append("like_complexity;");
		fw.append("prefer_little_thought;");
		fw.append("how_is_unimportant;");
		fw.append("not_like_noverlty;");
		fw.append("familiar_bpmn;");
		
		fw.append("other_techniques;");
		fw.append("most_familiar_technique;");
		fw.append("thought_to_top;");
		fw.append("like_abstraction;");
		fw.append("not_like_thinking;");
		fw.append("current_profession;");
		fw.append("wf_config_id;"); 
	}
	private void writeDemographic(FileWriter fw, String key) throws IOException {
		fw.append(act_size_height.get(key) + ";");
		fw.append(act_size_width.get(key) + ";");
		fw.append(disp_height.get(key) + ";");
		fw.append(disp_width.get(key) + ";");
		fw.append(duration.get(key) + ";");
		fw.append(exp_wf_dur.get(key) + ";");
		
		fw.append(model_act.get(key) + ";");
		fw.append(years_bpmn.get(key) + ";");
		fw.append(models_analysed.get(key) + ";");
		fw.append(models_created.get(key) + ";");
		fw.append(formal_training.get(key) + ";");
		fw.append(self_education.get(key) + ";");
		fw.append(years_model.get(key) + ";");

		fw.append(mental_effort1.get(key) + ";");
		fw.append(mental_effort2.get(key) + ";");
		fw.append(know_preflight.get(key) + ";");
		fw.append(know_mortgage.get(key) + ";");
		fw.append(expert.get(key) + ";");
		fw.append(relief_satisfaction.get(key) + ";");
		fw.append(competent_use_bpmn.get(key) + ";");
		fw.append(competent_get_bpmn.get(key) + ";");
		fw.append(like_deliberating.get(key) + ";");
		fw.append(like_little_thought.get(key) + ";");
		fw.append(like_lot_thinking.get(key) + ";");
		fw.append(think_less.get(key) + ";");
		fw.append(like_puzzles.get(key) + ";");
		fw.append(like_small_projects.get(key) + ";");
		fw.append(like_noverlty.get(key) + ";");
		fw.append(avoid_thinking.get(key) + ";");
		fw.append(usually_deliberate.get(key) + ";");
		fw.append(like_intellectual.get(key) + ";");
		fw.append(like_complexity.get(key) + ";");
		fw.append(prefer_little_thought.get(key) + ";");
		fw.append(how_is_unimportant.get(key) + ";");
		fw.append(not_like_noverlty.get(key) + ";");
		fw.append(familiar_bpmn.get(key) + ";");
		
		fw.append(other_techniques.get(key) + ";");
		fw.append(most_familiar_technique.get(key) + ";");
		fw.append(thought_to_top.get(key) + ";");
		fw.append(like_abstraction.get(key) + ";");
		fw.append(not_like_thinking.get(key) + ";");
		fw.append(current_profession.get(key) + ";");
		fw.append(wf_config_id.get(key) + ";"); 
	}
	
	//in relation to
	private String irt(double a, double b) {
		return dblformat.format(Math.round((a * 10000.0) / b) / 100.0) + "%";
	}
	private String time(double sec) {
		return dblformat.format(sec);
	}
	private String round(double avg) {
		return dblformat.format(Math.round(avg * 100.0) / 100.0);
	}
	
	private int numEvent(String key) {
		return numCreate(key) + 
		       numMove(key) + 
		       numDelete(key) + 
		       numRename(key);
	}
	
	private int numCreate(String key) {
		return num_create_activity.get(key) + 
		       num_create_and.get(key) + 
		       num_create_xor.get(key) +
		       num_create_start_event.get(key) +
		       num_create_end_event.get(key) +
		       num_create_edge.get(key) +
		       num_reconnect_edge.get(key);
	}
	private int numMove(String key) {
		return num_move_activity.get(key) + 
	           num_move_and.get(key) + 
	           num_move_xor.get(key) +
	           num_move_start_event.get(key) +
	           num_move_end_event.get(key) +
	           num_create_edge_bendpoint.get(key) + 
	           num_move_edge_bendpoint.get(key) + 
	           num_delete_edge_bendpoint.get(key) + 
	           num_move_edge_label.get(key);
	}
	private int numDelete(String key) {
		return num_delete_activity.get(key) + 
		       num_delete_and.get(key) + 
		       num_delete_xor.get(key) +
		       num_delete_start_event.get(key) +
		       num_delete_end_event.get(key) +
		       num_delete_edge.get(key) +
		       num_reconnect_edge.get(key);
	}
	private int numRename(String key) {
		return num_name_activity.get(key) +
		       num_name_edge.get(key) + 
		       num_rename_activity.get(key) + 
		       num_rename_edge.get(key);
	}
	
	private CountHashMap getNumMap(String activity) {
		if (activity.equalsIgnoreCase("CREATE_ACTIVITY"))
			return num_create_activity;
		else if (activity.equalsIgnoreCase("CREATE_AND"))
			return num_create_and;
		else if (activity.equalsIgnoreCase("CREATE_XOR"))
			return num_create_xor;
		else if (activity.equalsIgnoreCase("CREATE_START_EVENT"))
			return num_create_start_event;
		else if (activity.equalsIgnoreCase("CREATE_END_EVENT"))
			return num_create_end_event;
		else if (activity.equalsIgnoreCase("CREATE_EDGE"))
			return num_create_edge;
		else if (activity.equalsIgnoreCase("CREATE_EDGE_BENDPOINT"))
			return num_create_edge_bendpoint;
		else if (activity.equalsIgnoreCase("MOVE_ACTIVITY"))
			return num_move_activity;
		else if (activity.equalsIgnoreCase("MOVE_AND"))
			return num_move_and;
		else if (activity.equalsIgnoreCase("MOVE_XOR"))
			return num_move_xor;
		else if (activity.equalsIgnoreCase("MOVE_START_EVENT"))
			return num_move_start_event;
		else if (activity.equalsIgnoreCase("MOVE_END_EVENT"))
			return num_move_end_event;
		else if (activity.equalsIgnoreCase("MOVE_EDGE_BENDPOINT"))
			return num_move_edge_bendpoint;
		else if (activity.equalsIgnoreCase("MOVE_EDGE_LABEL"))
			return num_move_edge_label;
		else if (activity.equalsIgnoreCase("DELETE_ACTIVITY"))
			return num_delete_activity;
		else if (activity.equalsIgnoreCase("DELETE_AND"))
			return num_delete_and;
		else if (activity.equalsIgnoreCase("DELETE_XOR"))
			return num_delete_xor;
		else if (activity.equalsIgnoreCase("DELETE_START_EVENT"))
			return num_delete_start_event;
		else if (activity.equalsIgnoreCase("DELETE_END_EVENT"))
			return num_delete_end_event;
		else if (activity.equalsIgnoreCase("DELETE_EDGE"))
			return num_delete_edge;
		else if (activity.equalsIgnoreCase("DELETE_EDGE_BENDPOINT"))
			return num_delete_edge_bendpoint;
		else if (activity.equalsIgnoreCase("RECONNECT_EDGE"))
			return num_reconnect_edge;
		else if (activity.equalsIgnoreCase("NAME_ACTIVITY"))
			return num_name_activity;
		else if (activity.equalsIgnoreCase("NAME_EDGE"))
			return num_name_edge;
		else if (activity.equalsIgnoreCase("RENAME_ACTIVITY"))
			return num_rename_activity;
		else if (activity.equalsIgnoreCase("RENAME_EDGE"))
			return num_rename_edge;
		else
			return null;
	}
	
	private boolean isCreate(String activity) {
		return (activity.toUpperCase().startsWith("CREATE_") && !isEdgeBendpoint(activity)) || isReconnect(activity);
	}
	private boolean isMove(String activity) {
		return activity.toUpperCase().startsWith("MOVE_") || isEdgeBendpoint(activity);
	}
	private boolean isDelete(String activity) {
		return (activity.toUpperCase().startsWith("DELETE_") && !isEdgeBendpoint(activity)) || isReconnect(activity);
	}
	private boolean isActivity(String activity) {
		return activity.toUpperCase().endsWith("_ACTIVITY");
	}
	private boolean isAnd(String activity) {
		return activity.toUpperCase().endsWith("_AND");
	}
	private boolean isXor(String activity) {
		return activity.toUpperCase().endsWith("_XOR");
	}
	private boolean isGateway(String activity) {
		return isAnd(activity) || isXor(activity);
	}
	private boolean isStart(String activity) {
		return activity.toUpperCase().endsWith("_START_EVENT");
	}
	private boolean isEnd(String activity) {
		return activity.toUpperCase().endsWith("_END_EVENT");
	}
	private boolean isStartEnd(String activity) {
		return isStart(activity) || isEnd(activity);
	}
	private boolean isEdge(String activity) {
		return activity.toUpperCase().endsWith("_EDGE");
	}
	private boolean isEdgeBendpoint(String activity) {
		return activity.toUpperCase().endsWith("_EDGE_BENDPOINT");
	}
	private boolean isReconnect(String activity) {
		return activity.toUpperCase().startsWith("RECONNECT_");
	}
	private boolean isEdgeLabel(String activity) {
		return activity.toUpperCase().endsWith("_EDGE_LABEL");
	}
	private String getID(XEvent event) {
		try {
			return event.getAttributes().get("id").toString();
		} catch (Exception ex) {
			return "";
		}
	}
}