package org.processmining.plugins.ppm.statistics;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.math.RoundingMode;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileNameExtensionFilter;

import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.plugins.ppm.ExportPPMCharts;
import org.processmining.plugins.ppm.ppmchart.model.BPMNModel;
import org.processmining.plugins.ppm.ppmchart.model.ChunkModel;
import org.processmining.plugins.ppm.ppmchart.model.DurationModel;
import org.processmining.plugins.ppm.ppmchart.ui.PPMChartPanel;
import org.processmining.plugins.ppm.statistics.ordermodel.OrderModel;
import org.processmining.plugins.simplelogoperations.XLogFunctions;

public class CalculatePPMStatistics {
	private static final int OPERATION_TYPE = 5;
	private static final int ELEMENT_TYPE = 1;
	private static final int LOOKBACK_WINDOW = 1;
	private static final boolean SIMUL_OPERATIONS_IS_ONE = false;
	
	private static final String ACT_SIZE_HEIGHT = "activity_size_height";
	private static final String ACT_SIZE_WIDTH = "activity_size_width";
	private static final String DISP_HEIGHT = "display_resolution_height";
	private static final String DISP_WIDTH = "display_resolution_width";
	private static final String DURATION = "durationInSeconds";
	private static final String EXP_WF_DUR= "experimental workflow activity duration";
	
	private static final String MODEL_ACT = "How many activities did all these models have on average?";
	private static final String YEARS_BPMN = "How many months ago did you start using BPMN? (The first version of BPMN stems from May 2004, i.e. 60 months until May 2009)";
	private static final String MODELS_ANALYSED = "How many process models have you analyzed or read within the last 12 months? (A year has about 250 work days. In case you read one model per day, this would sum up to 250 models per year)";
	private static final String MODELS_CREATED = "How many process models have you created or edited within the last 12 months?";
	private static final String FORMAL_TRAINING = "How many work days of formal training on process modeling have you received within the last 12 months? (This includes e.g. university lectures, certification courses, training courses. 15 weeks of a 90 minutes university lecture is roughly 3 work days)_duration";
	private static final String SELF_EDUCATION = "How many work days of self education have you made within the last 12 months? (This includes e.g. learning-by-doing, learning-on-the-fly, self-study of textbooks or specifications) _duration";
	private static final String YEARS_MODEL = "How many years ago did you start process modeling?";
	private static final String MENTAL_EFFORT1 = "How would you assess the mental effort for completing the first modeling task?";
	private static final String MENTAL_EFFORT2= "How would you assess the mental effort for completing the second modeling task?";
	private static final String KNOW_PREFLIGHT = "I am familiar with Pre-Flight Processes.";
	private static final String KNOW_MORTGAGE = "I am very familiar with mortgage processes.";
	private static final String EXPERT = "I consider myself being a process modeling expert.";
	private static final String RELIEF_SATISFACTION = "I feel relief rather than satisfaction after completing a task that required a lot of mental effort.";
	private static final String COMPETENT_USE_BPMN = "I feel very competent in using the BPMN for process modeling.";
	private static final String COMPETENT_GET_BPMN = "I feel very confident in understanding process models created with the BPMN.";
	private static final String LIKE_DELIBERATING = "I find satisfaction in deliberating hard and for long hours.";
	private static final String LIKE_LITTLE_THOUGHT = "I like tasks that require little thought once I&apos;ve learned them.";
	private static final String LIKE_LOT_THINKING = "I like to have the responsibility of handling a situation that requires a lot of thinking.";
	private static final String THINK_LESS = "I only think as hard as I have to.";
	private static final String LIKE_PUZZLES = "I prefer my Iife to be filled with puzzles that I must solve.";
	private static final String LIKE_SMALL_PROJECTS = "I prefer to think about small, daily projects to long-term ones.";
	private static final String LIKE_NOVELTY = "I really enjoy a task that involves coming up with new solutions to problems.";
	private static final String AVOID_THINKING = "I try to anticipate and avoid situations where there is likely chance I will have to think in depth about something.";
	private static final String USUALLY_DELIBERATE = "I usually end up deliberating about issues even when they do not affect me personally.";
	private static final String LIKE_INTELLECTUAL = "I would prefer a task that is intellectual, difficult, and important to one that is somewhat important but does not require much thought.";
	private static final String LIKE_COMPLEXITY = "I would prefer complex to simple problems.";
	private static final String PREFER_LITTLE_THOUGHT = "I would rather do something that requires little thought than something that is sure to challenge my thinking abilities.";
	private static final String HOW_IS_UNIMPORTANT = "It&apos;s enough for me that something gets the job done; I don&apos;t care how or why it works.";
	private static final String NOT_LIKE_NOVERLTY = "Learning new ways to think doesn&apos;t excite me very much.";
	private static final String FAMILIAR_BPMN = "Overall, I am very familiar with the BPMN.";
	private static final String OTHER_TECHNIQUES = "Please specify further modeling techniques you are familiar with (e.g. BPMN, Petri nets). If those techniques are closely associated to tools (e.g., Mavim, BizzDesigner), please specify those tools.";
	private static final String MOST_FAMILIAR_TECHNIQUE = "Please specify the modeling technique you are most familiar with (e.g. BPMN, Petri nets). If that technique is closely associated to a modeling tool (e.g., Mavim, BizzDesigner), please specify that tool.";
	private static final String THOUGHT_TO_TOP = "The idea of relying on thought to make my way to the top appeals to me.";
	private static final String LIKE_ABSTRACTION = "The notion of thinking abstractly is appealing to me.";
	private static final String NOT_LIKE_THINKING = "Thinking is not my idea of fun.";
	private static final String CURRENT_PROFESSION = "What is your current profession (e.g., student, consultant)?";
	private static final String WF_CONFIG_ID = "workflow_configuration_id";
	
	private static NumberFormat dblformat = NumberFormat.getNumberInstance();
		
	private ArrayList<String> models = new ArrayList<String>();
	private HashMap<String,String> task = new HashMap<String, String>();
	
	private CountHashMap act_size_height = new CountHashMap();
	private CountHashMap act_size_width = new CountHashMap();
	private CountHashMap disp_height = new CountHashMap();
	private CountHashMap disp_width = new CountHashMap();
	private CountHashMap duration = new CountHashMap();
	private CountHashMap exp_wf_dur = new CountHashMap();
	
	private CountHashMap model_act = new CountHashMap();
	private CountHashMap years_bpmn = new CountHashMap();
	private CountHashMap models_analysed = new CountHashMap();
	private CountHashMap models_created = new CountHashMap();
	private CountHashMap formal_training = new CountHashMap();
	private CountHashMap self_education = new CountHashMap();
	private CountHashMap years_model = new CountHashMap();

	private HashMap<String,String> mental_effort1 = new HashMap<String, String>();
	private HashMap<String,String> mental_effort2 = new HashMap<String, String>();
	private HashMap<String,String> know_preflight = new HashMap<String, String>();
	private HashMap<String,String> know_mortgage = new HashMap<String, String>();
	private HashMap<String,String> expert = new HashMap<String, String>();
	private HashMap<String,String> relief_satisfaction = new HashMap<String, String>();
	private HashMap<String,String> competent_use_bpmn = new HashMap<String, String>();
	private HashMap<String,String> competent_get_bpmn = new HashMap<String, String>();
	private HashMap<String,String> like_deliberating = new HashMap<String, String>();
	private HashMap<String,String> like_little_thought = new HashMap<String, String>();
	private HashMap<String,String> like_lot_thinking = new HashMap<String, String>();
	private HashMap<String,String> think_less = new HashMap<String, String>();
	private HashMap<String,String> like_puzzles = new HashMap<String, String>();
	private HashMap<String,String> like_small_projects = new HashMap<String, String>();
	private HashMap<String,String> like_noverlty = new HashMap<String, String>();
	private HashMap<String,String> avoid_thinking = new HashMap<String, String>();
	private HashMap<String,String> usually_deliberate = new HashMap<String, String>();
	private HashMap<String,String> like_intellectual = new HashMap<String, String>();
	private HashMap<String,String> like_complexity = new HashMap<String, String>();
	private HashMap<String,String> prefer_little_thought = new HashMap<String, String>();
	private HashMap<String,String> how_is_unimportant = new HashMap<String, String>();
	private HashMap<String,String> not_like_noverlty = new HashMap<String, String>();
	private HashMap<String,String> familiar_bpmn = new HashMap<String, String>();
	
	private HashMap<String,String> other_techniques = new HashMap<String, String>();
	private HashMap<String,String> most_familiar_technique = new HashMap<String, String>();
	private HashMap<String,String> thought_to_top = new HashMap<String, String>();
	private HashMap<String,String> like_abstraction = new HashMap<String, String>();
	private HashMap<String,String> not_like_thinking = new HashMap<String, String>();
	private HashMap<String,String> current_profession = new HashMap<String, String>();
	private HashMap<String,String> wf_config_id = new HashMap<String, String>(); 
	
	private DoubleHashMap aspect_orientedness = new DoubleHashMap();
	private DoubleHashMap flow_orientedness = new DoubleHashMap();
	private DoubleHashMap x_values = new DoubleHashMap();
	private DoubleHashMap y_values = new DoubleHashMap();
	private CountHashMap num_create_activity = new CountHashMap();
	private CountHashMap num_create_and = new CountHashMap();
	private CountHashMap num_create_xor = new CountHashMap();
	private CountHashMap num_create_start_event = new CountHashMap();
	private CountHashMap num_create_end_event = new CountHashMap();
	private CountHashMap num_create_edge = new CountHashMap();
	private CountHashMap num_create_edge_bendpoint = new CountHashMap();
	private CountHashMap num_move_activity = new CountHashMap();
	private CountHashMap num_move_and = new CountHashMap();
	private CountHashMap num_move_xor = new CountHashMap();
	private CountHashMap num_move_start_event = new CountHashMap();
	private CountHashMap num_move_end_event = new CountHashMap();
	private CountHashMap num_move_edge_bendpoint = new CountHashMap();
	private CountHashMap num_move_edge_label = new CountHashMap();
	private CountHashMap num_delete_activity = new CountHashMap();
	private CountHashMap num_delete_and = new CountHashMap();
	private CountHashMap num_delete_xor = new CountHashMap();
	private CountHashMap num_delete_start_event = new CountHashMap();
	private CountHashMap num_delete_end_event = new CountHashMap();
	private CountHashMap num_delete_edge = new CountHashMap();
	private CountHashMap num_delete_edge_bendpoint = new CountHashMap();
	private CountHashMap num_reconnect_edge = new CountHashMap();
	private CountHashMap num_name_activity = new CountHashMap();
	private CountHashMap num_name_edge = new CountHashMap();
	private CountHashMap num_rename_activity = new CountHashMap();
	private CountHashMap num_rename_edge = new CountHashMap();
	
	private CountHashMap num_create_activity_before_create_end_inclgw = new CountHashMap();
	private CountHashMap num_create_activity_before_create_end_exclgw = new CountHashMap();
	private CountHashMap num_create_before_create_end_inclgw = new CountHashMap();
	private CountHashMap num_create_before_create_end_exclgw = new CountHashMap();
	private CountHashMap num_move_activity_before_create_end_inclgw = new CountHashMap();
	private CountHashMap num_move_activity_before_create_end_exclgw = new CountHashMap();
	private CountHashMap num_move_activity_before_last_update_end_inclgw = new CountHashMap();
	private CountHashMap num_move_activity_before_last_update_end_exclgw = new CountHashMap();
	
	private CountHashMap num_element_with_moves = new CountHashMap();
	private DoubleHashMap avg_num_move_move_element = new DoubleHashMap();
	private CountHashMap max_num_move_move_element = new CountHashMap();
	private DoubleHashMap avg_num_update_one_element = new DoubleHashMap();
	private CountHashMap max_num_update_one_element = new CountHashMap();
	private DoubleHashMap avg_num_consecutive_create = new DoubleHashMap();
	private CountHashMap max_num_consecutive_create = new CountHashMap();
	private DoubleHashMap avg_num_consecutive_move = new DoubleHashMap();
	private CountHashMap max_num_consecutive_move = new CountHashMap();
	private DoubleHashMap avg_num_consecutive_delete = new DoubleHashMap();
	private CountHashMap max_num_consecutive_delete = new CountHashMap();
	
	private DoubleHashMap tot_time = new DoubleHashMap();
	private DoubleHashMap tot_create_time = new DoubleHashMap();
	private DoubleHashMap avg_gap = new DoubleHashMap();
	private DoubleHashMap max_gap = new DoubleHashMap();
	private CountHashMap num_gap5s = new CountHashMap();
	private CountHashMap num_gap10s = new CountHashMap();
	private CountHashMap num_gap15s = new CountHashMap();
	private CountHashMap num_gap20s = new CountHashMap();
	private CountHashMap num_gap25s = new CountHashMap();
	private CountHashMap num_gap30s = new CountHashMap();
	private CountHashMap num_gap31s = new CountHashMap();

	private DoubleHashMap avg_intermediate_update_gateway = new DoubleHashMap();
	private CountHashMap max_intermediate_update_gateway = new CountHashMap();
	private DoubleHashMap avg_intermediate_create_gateway = new DoubleHashMap();
	private CountHashMap max_intermediate_create_gateway = new CountHashMap();
	private DoubleHashMap avg_intermediate_create_block = new DoubleHashMap();
	private CountHashMap max_intermediate_create_block = new CountHashMap(); 
	private DoubleHashMap avg_intermediate_update = new DoubleHashMap();
	private CountHashMap max_intermediate_update = new CountHashMap();
	private DoubleHashMap avg_intermediate_create = new DoubleHashMap();
	private CountHashMap max_intermediate_create = new CountHashMap(); 
	
	private CountHashMap num_block_as_a_whole = new CountHashMap();
	private CountHashMap num_blocks_in_pieces = new CountHashMap();
	private CountHashMap max_simul_block = new CountHashMap();
	
	private CountHashMap num_elements_between_create_nodes = new CountHashMap();
	private CountHashMap num_ops_between_create_nodes = new CountHashMap();
	private CountHashMap num_elements_between_ops = new CountHashMap();
	private CountHashMap num_create_nodes = new CountHashMap();
	
	public CalculatePPMStatistics() {
		dblformat.setGroupingUsed(false);
		dblformat.setMaximumFractionDigits(5);
		dblformat.setRoundingMode(RoundingMode.HALF_UP);
	}
	
	@Plugin(name = "Calculate PPM Statistics",
			parameterLabels = { "PPM log" },
			returnLabels = { "File in filesystem" },
			returnTypes = { XLog.class },
			userAccessible = true,
			help = "Calculates statistics from a PPM log")
	@UITopiaVariant(affiliation = "Ghent University", 
			        author = "J. Claes", 
			        email = "jan.claes@ugent.be",
			        website = "http://www.janclaes.info")
	public XLog calculate(PluginContext context, XLog log) {
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(log.size());
		context.getProgress().setCaption("Analysing log...");
		context.getProgress().setIndeterminate(false);
		
		for (XTrace trace : log) {
			String key = XLogFunctions.getID(trace);
			
			models.add(key);
			task.put(key, getString(trace.getAttributes(), "PROCESS"));
			
			//calculateActivityStatistics(trace);
			//calculateAspectOrientedness(trace);
			//calculateFlowOrientedness(trace);
			//calculateBetweenStatistics(trace);
			calculateFitStatistics(trace);
			context.getProgress().inc();
		}
		String filename = exportToCsv();
		JOptionPane.showMessageDialog(null, "Statistics were written to " + filename);
		return log;
	}			
	
	private String getString(XAttributeMap map, String key) {
		if (!map.containsKey(key))
			return "";
		return ((XAttributeLiteral)map.get(key)).getValue();
	}

	private void calculateActivityStatistics(XTrace trace) {
		String key = XLogFunctions.getID(trace);
		for (XEvent event : trace) {
			String activity = XLogFunctions.getName(event);
			if (getNumMap(activity) != null) 
				getNumMap(activity).incr(key);
		}
	}
	
	private void calculateAspectOrientedness(XTrace trace) {
		if (SIMUL_OPERATIONS_IS_ONE)
			JOptionPane.showMessageDialog(null, "This operation is currently not yet supported");
		String key = XLogFunctions.getID(trace);
		ArrayList<String> benchmarks = new ArrayList<String>();
		int numEvent = 0;
		
		for (XEvent event : trace) {
			String activity = XLogFunctions.getName(event);
			if (!isEventOfInterest(activity))
				continue;
			if (benchmarks.isEmpty()) {
				benchmarks.add(activity);
				continue;
			}
			
			aspect_orientedness.add(key, getAspectDistance(activity, benchmarks));
			numEvent++;
			
			benchmarks.add(activity);
			if (benchmarks.size() > LOOKBACK_WINDOW)
				benchmarks.remove(0);
		}
		aspect_orientedness.put(key, aspect_orientedness.get(key) / numEvent);
	}
	private double getAspectDistance(String activity, ArrayList<String> benchmarks)
	{
		double minDistance = -1;
		for (String benchmark : benchmarks)
		{
			double distance = 0;
			if (!getOperationType(activity).equals(getOperationType(benchmark)))
				distance += OPERATION_TYPE;
			if (!getElementType(activity).equals(getElementType(benchmark)))
				distance += ELEMENT_TYPE;
			if (minDistance == -1 || distance < minDistance)
				minDistance = distance;
		}
		return minDistance;
	}
	private String getOperationType(String activity)
	{
		return activity.substring(0, activity.indexOf("_") - 1);
	}
	private String getElementType(String activity)
	{
		return activity.substring(activity.indexOf("_") + 1);
	}
	
	private void calculateFlowOrientedness(XTrace trace) {
		String key = XLogFunctions.getID(trace);
		XEvent prevEvent = null;
		int numEvent = 0;
		
		for (XEvent event : trace) {
			String activity = XLogFunctions.getName(event);
			if (!isEventOfInterest(activity))
				continue;
			
			register(event);
			if (prevEvent != null)
			{
				System.out.print(getFlowDistance(event, prevEvent) + " ");
				flow_orientedness.add(key, getFlowDistance(event, prevEvent));
				numEvent++;
			}
			prevEvent = event;
		}
		System.out.println(flow_orientedness.get(key) + "-" + numEvent(key) + "-" + numEvent);
		flow_orientedness.put(key, flow_orientedness.get(key));
	}
	private String getId(XEvent event)
	{
		if (event.getAttributes().containsKey("id"))
			return XLogFunctions.getStringValue(event.getAttributes().get("id"));
		else
			return "";
	}
	private long getX(XEvent event)
	{
		if (event.getAttributes().containsKey("x"))
			return XLogFunctions.getLongValue(event.getAttributes().get("x"));
		else
			return -1;
	}
	private long getY(XEvent event)
	{
		if (event.getAttributes().containsKey("y"))
			return XLogFunctions.getLongValue(event.getAttributes().get("y"));
		else
			return -1;
	}
	private void register(XEvent event)
	{
		XAttributeMap attMap = event.getAttributes();
		if (!attMap.containsKey("id") || !attMap.containsKey("x") || !attMap.containsKey("y"))
			return;
					
		x_values.put(getId(event), getX(event));
		y_values.put(getId(event), getY(event));
	}
	private double getXValue(XEvent event)
	{
		if (x_values.containsKey(getId(event)))
		{
			return x_values.get(getId(event));
		}
		else if (XLogFunctions.getName(event).equals("CREATE_EDGE"))
		{
			double x1 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
			double x2 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
			return (x1 + x2) / 2.0;
		}
		else if (XLogFunctions.getName(event).equals("RECONNECT_EDGE"))
		{
			if (!event.getAttributes().containsKey("source_node_id"))
				return x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
			else if (!event.getAttributes().containsKey("target_node_id"))
				return x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
			else if (!event.getAttributes().containsKey("old_node_id"))
			{
				double x1 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
				double x2 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
				return (x1 + x2) / 2.0;
			}
			else
			{
				double x2, x1 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("old_node_id")));
				if (XLogFunctions.getStringValue(event.getAttributes().get("old_node_type")).equalsIgnoreCase("Source"))
					x2 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
				else
					x2 = x_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
				return (x1 + x2) / 2.0;
			}
		}
		return -1;
	}
	private double getYValue(XEvent event)
	{
		if (y_values.containsKey(getId(event)))
		{
			return y_values.get(getId(event));
		}
		else if (XLogFunctions.getName(event).equals("CREATE_EDGE"))
		{
			double y1 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
			double y2 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
			return (y1 + y2) / 2.0;
		}
		else if (XLogFunctions.getName(event).equals("RECONNECT_EDGE"))
		{
			if (!event.getAttributes().containsKey("source_node_id"))
				return y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
			else if (!event.getAttributes().containsKey("target_node_id"))
				return y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
			else if (!event.getAttributes().containsKey("old_node_id"))
			{
				double y1 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
				double y2 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
				return (y1 + y2) / 2.0;
			}
			else
			{
				double y2, y1 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("old_node_id")));
				if (XLogFunctions.getStringValue(event.getAttributes().get("old_node_type")).equalsIgnoreCase("Source"))
					y2 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("source_node_id")));
				else
					y2 = y_values.get(XLogFunctions.getStringValue(event.getAttributes().get("target_node_id")));
				return (y1 + y2) / 2.0;
			}
		}
		return -1;
	}
	private double getFlowDistance(XEvent event1, XEvent event2)
	{
		try
		{
		double x1 = getXValue(event1);
		double y1 = getYValue(event1);
		double x2 = getXValue(event2);
		double y2 = getYValue(event2);
		return Math.sqrt(Math.pow((x2 - x1),2) + Math.pow((y2 - y1),2));
		}
		catch (NullPointerException ex) {
			ex.printStackTrace();
			return 0;
		}
	}
	
	private void calculateBetweenStatistics(XTrace trace) {
		String key = XLogFunctions.getID(trace);
		int prevEvent = -1;
		int prevCreateEvent = -1;
		int eventsToIgnore = 0;
		
		XEvent[] distances = null;
		try {
			DurationModel dm = DurationModel.getInstance(ExportPPMCharts.convertTrace(trace), PPMChartPanel.ST_DIST_START);
			distances = new XEvent[trace.size()]; //max trace.size() unique elements, probably far less
			int i = 0;
			for (XEvent event : trace) {
				if (!isCreate(XLogFunctions.getName(event)))
					continue;
				long dist = getDistance(dm, event);
				int j = i;
				while (j > 0 && getDistance(dm, distances[j-1]) > dist) {
					distances[j] = distances [j-1];
					j--;
				}
				distances[j] = event;
				i++;
			}	
			dm = null;
		} catch (Exception ex) {
			ex.printStackTrace();
		}

		
		for (int i=0; i<trace.size(); i++) {
			XEvent event = trace.get(i);
			String activity = XLogFunctions.getName(event);
			if (!isEventOfInterest(activity)) {
				eventsToIgnore++;
				continue;
			}
			
			if (isCreate(activity) && isNode(activity)) {
				num_create_nodes.incr(key);
				if (prevCreateEvent > -1) {
					num_ops_between_create_nodes.add(key,  i-prevCreateEvent-eventsToIgnore-1);
					num_elements_between_create_nodes.add(key, elementsBetween(distances, trace, prevCreateEvent, i));
				}
				prevCreateEvent = i;
				eventsToIgnore = 0;
			}
			if (prevEvent > -1) {
				num_elements_between_ops.add(key, elementsBetween(distances, trace, prevEvent, i));
			}
			prevEvent = i;
		}
	}
	private long getDistance(DurationModel dm, XEvent event) {
		long dist = Long.MAX_VALUE;
		if (event.getAttributes().containsKey("id"))
			dist = dm.getDistanceValue(getId(event));
		return dist;
	}
	private int elementsBetween(XEvent[] distances, XTrace trace, int index1, int index2) {
		try {
			return (int)Math.abs(find(distances, trace.get(index1)) - find(distances, trace.get(index2))) - 1;
		}
		catch (Exception ex) {
			ex.printStackTrace();
			return 0;
		}
	}
	private int find(XEvent[] distances, XEvent event) {
		for (int i=0; i<distances.length; i++)
			if (getId(distances[i]).equals(getId(event)))
				return i;
		return -1;
	}
	
	private DoubleHashMap avg_progress = new DoubleHashMap();
	private DoubleHashMap stddev_progress = new DoubleHashMap();
	private DoubleHashMap avg_distance = new DoubleHashMap();
	private DoubleHashMap max_distance = new DoubleHashMap();
	private DoubleHashMap avg_overlap = new DoubleHashMap();
	private DoubleHashMap max_overlap = new DoubleHashMap();
	private CountHashMap num_pauses = new CountHashMap();
	private CountHashMap num_deletes = new CountHashMap();
	private CountHashMap num_moves = new CountHashMap();
	private CountHashMap num_reconnects = new CountHashMap();
	private CountHashMap num_alterations = new CountHashMap();
	private CountHashMap difference = new CountHashMap();

	private void calculateFitStatistics(XTrace trace) {
		String key = XLogFunctions.getID(trace);
		int numOperations = 0;
		
		XLog traceLog = ExportPPMCharts.convertTrace(trace);
		
		//progress
		long timeOfFirstOperation = -1;
		long timeOfLastOperation = XLogFunctions.getTime(trace.get(trace.size()-1)).getTime();
		double numMillisInTimeSlot = 3.0 * 60 * 1000; // 3 min
		int numTimeSlots = 0;
		double[] timeSlots = new double[0];
		
		//distance
		BPMNModel distanceModel = new BPMNModel(traceLog, true, false, true);
		double[] distances = new double[trace.size()-1];
		int distCounter = 0;
		
		//pauses
		double[] gaps = new double[trace.size()-1];
		int gapCounter = 0;
		
		XEvent prevEvent = null;
		for (XEvent event : trace) {
			String activity = XLogFunctions.getName(event);
			if (!isEventOfInterest(activity))
				continue;

			//progress
			long timeOfOperation = XLogFunctions.getTime(event).getTime();
			if (timeOfFirstOperation == -1) {
				timeOfFirstOperation = timeOfOperation;
				numTimeSlots = (int)Math.ceil((timeOfLastOperation - timeOfFirstOperation) / numMillisInTimeSlot);
				timeSlots = new double[numTimeSlots];
				for (int i=0; i<numTimeSlots; i++)
					timeSlots[i] = 0;
			}
			int timeSlot = (int)Math.floor((timeOfOperation - timeOfFirstOperation) / numMillisInTimeSlot);
			if (timeOfOperation == timeOfLastOperation)
				timeSlot--;

			timeSlots[timeSlot]++;
			
			//distances
			if (prevEvent != null)
				distances[distCounter++] = distanceModel.getDistance(getId(prevEvent), getId(event));
			
			//pauses
			if (prevEvent != null)
				gaps[gapCounter++] = (XLogFunctions.getTime(event).getTime() - XLogFunctions.getTime(prevEvent).getTime()) / 1000.0; // 1 sec
			
			//alterations
			if (isReconnect(activity)) {
				num_reconnects.incr(key);
				num_alterations.incr(key);
			}
			else if (isDelete(activity)) { //attention: deletes excluding reconnects because of else (don't change order)
				num_deletes.incr(key);
				num_alterations.incr(key);
			}
			else if (isMove(activity)) {
				num_moves.incr(key);
				num_alterations.incr(key);
			}
			
			prevEvent = event;
			numOperations++;
		}
		
		//progress
		avg_progress.add(key, trimmedAverage(timeSlots));
		stddev_progress.add(key, trimmedStddev(timeSlots));
		
		//distances
		avg_distance.add(key, average(distances, distCounter - 1));
		max_distance.add(key, maximum(distances, distCounter - 1));
		
		//overlap
		ChunkModel chunkModel = new ChunkModel(traceLog, 0, 0, 0, 0);
		double[] overlap = new double[chunkModel.getBlockElements().size()];
		for (int i=0; i<chunkModel.getBlockElements().size(); i++)
			overlap[i] = chunkModel.getBlockElements().get(i).getDistance();
		avg_overlap.add(key, average(overlap));
		max_overlap.add(key, maximum(overlap));
		
		//pauses (pause = gap above threshold)
		double threshold = average(gaps, gapCounter - 1) + stddev(gaps, gapCounter - 1);
		int numAboveThreshold = 0;
		for (double gap : gaps)
			if (gap > threshold)
				numAboveThreshold++;
		num_pauses.put(key, numAboveThreshold);

		//difference
		OrderModel orderModel = new OrderModel(traceLog);
		difference.add(key, orderModel.getDifference());
		

	}
	private double[] replace(double[] array, double val1, double val2) {
		for (int i=0; i<array.length; i++)
			if (array[i] == val1)
				array[i] = val2;
		return array;
	}
	private double trimmedAverage(double[] values) {
		int num = values.length - 1;
		while (num >= 0 && values[num] == 0)
			num--;
		return average(values, num + 1);
	}
	private double average(double[] values) {
		return average(values, values.length);
	}
	private double average(double[] values, int n) {
		int num = Math.min(values.length, n);
		double sum = 0, cnt = 0;
		for (int i=0; i<num; i++) {
			if (values[i] == Double.MAX_VALUE || values[i] == Double.NEGATIVE_INFINITY || values[i] == Double.POSITIVE_INFINITY)
				continue;
			sum += values[i];
			cnt++;
		}
		return sum / cnt;
	}
	private double trimmedStddev(double[] values) {
		int num = values.length - 1;
		while (num >= 0 && values[num] == 0)
			num--;
		return stddev(values, num + 1);
	}
	private double stddev(double[] values) {
		return stddev(values, values.length);
	}
	private double stddev(double[] values, int n) {
		int num = Math.min(values.length, n);
		double stddev = 0, cnt = 0;
		double mean = average(values, n);
		for (int i=0; i<num; i++) {
			if (values[i] == Double.MAX_VALUE || values[i] == Double.NEGATIVE_INFINITY || values[i] == Double.POSITIVE_INFINITY)
				continue;
			stddev += Math.pow((mean - values[i]), 2);
			cnt++;
		}
		stddev /= cnt;
		return Math.sqrt(stddev);
	}
	private double trimmedMaximum(double[] values) {
		int num = values.length - 1;
		while (num >= 0 && values[num] == 0)
			num--;
		return maximum(values, num + 1);
	}
	private double maximum(double[] values) {
		return maximum(values, values.length);
	}
	private double maximum(double[] values, int n) {
		int num = Math.min(values.length, n);
		double max = values[0];
		for (int i=0; i<num; i++) {
			if (values[i] == Double.MAX_VALUE || values[i] == Double.NEGATIVE_INFINITY || values[i] == Double.POSITIVE_INFINITY)
				continue;
			if (values[i] > max)
				max = values[i];
		}
		return max;
	}

	
	private String exportToCsv() {
		JFileChooser fc = new JFileChooser();
		fc.setFileFilter(new FileNameExtensionFilter("*.csv files", "csv"));
		if (fc.showSaveDialog(null) == JFileChooser.APPROVE_OPTION) {
			try {
				//TODO ask overwrite if file exists
				File file = fc.getSelectedFile();
				FileWriter fw = new FileWriter(file);
				fw.write("CaseNum;Task;");
				
				//writeAspectOrientednessHeading(fw);
				//writeFlowOrientednessHeading(fw);
				//writeBetweenStatisticsHeading(fw);
				writeFitStatisticsHeading(fw);
				
				fw.append("\n");
				
				for (String key : models) {
					fw.append(key + ";");
					fw.append(task.get(key) + ";");
					
					//writeAspectOrientedness(fw, key);
					//writeFlowOrientedness(fw, key);
					//writeBetweenStatistics(fw, key);
					writeFitStatistics(fw, key);
					
					fw.append("\n");
				}
				fw.close();
				return file.getPath();
			}
			catch (Exception ex) {
				ex.printStackTrace();
				JOptionPane.showMessageDialog(null, "Error trying to save results to file", "Error", JOptionPane.ERROR_MESSAGE);
			}
		}
		return "ERROR";
	}
	
	private void writeAspectOrientednessHeading(FileWriter fw) throws IOException {
		fw.append("aspect-orientedness;");
	}
	private void writeAspectOrientedness(FileWriter fw, String key) throws IOException {
		fw.append(round(aspect_orientedness.get(key)) + ";");
	}
	private void writeFlowOrientednessHeading(FileWriter fw) throws IOException {
		fw.append("flow-orientedness;");
	}
	private void writeFlowOrientedness(FileWriter fw, String key) throws IOException {
		fw.append(round(flow_orientedness.get(key)) + ";");
	}
	private void writeBetweenStatisticsHeading(FileWriter fw) throws IOException {
		fw.append("avg_elements_between_create_nodes;avg_ops_between_create_nodes;avg_elements_between_ops;");
		fw.append("num_elements_between_create_nodes;num_ops_between_create_nodes;");
		fw.append("num_elements_between_ops;num_create_nodes;numEvent;"); 
	}
	private void writeBetweenStatistics(FileWriter fw, String key) throws IOException {
		fw.append(round(num_elements_between_create_nodes.get(key) / (1.0 * (num_create_nodes.get(key) - 1))) + ";");
		fw.append(round(num_ops_between_create_nodes.get(key) / (1.0 * (num_create_nodes.get(key) - 1))) + ";");
		fw.append(round(num_elements_between_ops.get(key) / (1.0 * (numEvent(key) - 1))) + ";");
		fw.append(num_elements_between_create_nodes.get(key) + ";");
		fw.append(num_ops_between_create_nodes.get(key) + ";");
		fw.append(num_elements_between_ops.get(key) + ";");
		fw.append(num_create_nodes.get(key) + ";");
		fw.append(numEvent(key) + ";");
	}
	private void writeFitStatisticsHeading(FileWriter fw) throws IOException {
		fw.append("avg_progress;stddev_progress;avg_distance;max_distance;avg_overlap;max_overlap;num_pauses;num_deletes;num_moves;num_reconnects;num_alterations;diffference;");
	}
	private void writeFitStatistics(FileWriter fw, String key) throws IOException {
		fw.append(format(avg_progress.get(key)) + ";");
		fw.append(format(stddev_progress.get(key)) + ";");
		fw.append(format(avg_distance.get(key)) + ";");
		fw.append(format(max_distance.get(key)) + ";");
		fw.append(format(avg_overlap.get(key)) + ";");
		fw.append(format(max_overlap.get(key)) + ";");
		fw.append(num_pauses.get(key) + ";");
		fw.append(num_deletes.get(key) + ";");
		fw.append(num_moves.get(key) + ";");
		fw.append(num_reconnects.get(key) + ";");
		fw.append(num_alterations.get(key) + ";");
		fw.append(difference.get(key) + ";");
	}

	
	//in relation to
	private String irt(double a, double b) {
		return format(Math.round((a * 10000.0) / b) / 100.0) + "%";
	}
	private String time(double sec) {
		return format(sec);
	}
	private String round(double avg) {
		return format(Math.round(avg * 100.0) / 100.0);
	}
	private String format(double d) {
		return dblformat.format(d).replace(".", ",");
	}
	
	private int numEvent(String key) {
		return numCreate(key) + 
		       numMove(key) + 
		       numDelete(key) + 
		       numRename(key);
	}
	
	private int numCreate(String key) {
		return num_create_activity.get(key) + 
		       num_create_and.get(key) + 
		       num_create_xor.get(key) +
		       num_create_start_event.get(key) +
		       num_create_end_event.get(key) +
		       num_create_edge.get(key) +
		       num_reconnect_edge.get(key);
	}
	private int numMove(String key) {
		return num_move_activity.get(key) + 
	           num_move_and.get(key) + 
	           num_move_xor.get(key) +
	           num_move_start_event.get(key) +
	           num_move_end_event.get(key) +
	           num_create_edge_bendpoint.get(key) + 
	           num_move_edge_bendpoint.get(key) + 
	           num_delete_edge_bendpoint.get(key) + 
	           num_move_edge_label.get(key);
	}
	private int numDelete(String key) {
		return num_delete_activity.get(key) + 
		       num_delete_and.get(key) + 
		       num_delete_xor.get(key) +
		       num_delete_start_event.get(key) +
		       num_delete_end_event.get(key) +
		       num_delete_edge.get(key) +
		       num_reconnect_edge.get(key);
	}
	private int numRename(String key) {
		return num_name_activity.get(key) +
		       num_name_edge.get(key) + 
		       num_rename_activity.get(key) + 
		       num_rename_edge.get(key);
	}
	
	private CountHashMap getNumMap(String activity) {
		if (activity.equalsIgnoreCase("CREATE_ACTIVITY"))
			return num_create_activity;
		else if (activity.equalsIgnoreCase("CREATE_AND"))
			return num_create_and;
		else if (activity.equalsIgnoreCase("CREATE_XOR"))
			return num_create_xor;
		else if (activity.equalsIgnoreCase("CREATE_START_EVENT"))
			return num_create_start_event;
		else if (activity.equalsIgnoreCase("CREATE_END_EVENT"))
			return num_create_end_event;
		else if (activity.equalsIgnoreCase("CREATE_EDGE"))
			return num_create_edge;
		else if (activity.equalsIgnoreCase("CREATE_EDGE_BENDPOINT"))
			return num_create_edge_bendpoint;
		else if (activity.equalsIgnoreCase("MOVE_ACTIVITY"))
			return num_move_activity;
		else if (activity.equalsIgnoreCase("MOVE_AND"))
			return num_move_and;
		else if (activity.equalsIgnoreCase("MOVE_XOR"))
			return num_move_xor;
		else if (activity.equalsIgnoreCase("MOVE_START_EVENT"))
			return num_move_start_event;
		else if (activity.equalsIgnoreCase("MOVE_END_EVENT"))
			return num_move_end_event;
		else if (activity.equalsIgnoreCase("MOVE_EDGE_BENDPOINT"))
			return num_move_edge_bendpoint;
		else if (activity.equalsIgnoreCase("MOVE_EDGE_LABEL"))
			return num_move_edge_label;
		else if (activity.equalsIgnoreCase("DELETE_ACTIVITY"))
			return num_delete_activity;
		else if (activity.equalsIgnoreCase("DELETE_AND"))
			return num_delete_and;
		else if (activity.equalsIgnoreCase("DELETE_XOR"))
			return num_delete_xor;
		else if (activity.equalsIgnoreCase("DELETE_START_EVENT"))
			return num_delete_start_event;
		else if (activity.equalsIgnoreCase("DELETE_END_EVENT"))
			return num_delete_end_event;
		else if (activity.equalsIgnoreCase("DELETE_EDGE"))
			return num_delete_edge;
		else if (activity.equalsIgnoreCase("DELETE_EDGE_BENDPOINT"))
			return num_delete_edge_bendpoint;
		else if (activity.equalsIgnoreCase("RECONNECT_EDGE"))
			return num_reconnect_edge;
		else if (activity.equalsIgnoreCase("NAME_ACTIVITY"))
			return num_name_activity;
		else if (activity.equalsIgnoreCase("NAME_EDGE"))
			return num_name_edge;
		else if (activity.equalsIgnoreCase("RENAME_ACTIVITY"))
			return num_rename_activity;
		else if (activity.equalsIgnoreCase("RENAME_EDGE"))
			return num_rename_edge;
		else
			return null;
	}
	
	private boolean isCreate(String activity) {
		return (activity.toUpperCase().startsWith("CREATE_") && !isEdgeBendpoint(activity)) || isReconnect(activity);
	}
	private boolean isMove(String activity) {
		return activity.toUpperCase().startsWith("MOVE_") || isEdgeBendpoint(activity);
	}
	private boolean isDelete(String activity) {
		return (activity.toUpperCase().startsWith("DELETE_") && !isEdgeBendpoint(activity)) || isReconnect(activity);
	}
	private boolean isActivity(String activity) {
		return activity.toUpperCase().endsWith("_ACTIVITY");
	}
	private boolean isAnd(String activity) {
		return activity.toUpperCase().endsWith("_AND");
	}
	private boolean isXor(String activity) {
		return activity.toUpperCase().endsWith("_XOR");
	}
	private boolean isGateway(String activity) {
		return isAnd(activity) || isXor(activity);
	}
	private boolean isStart(String activity) {
		return activity.toUpperCase().endsWith("_START_EVENT");
	}
	private boolean isEnd(String activity) {
		return activity.toUpperCase().endsWith("_END_EVENT");
	}
	private boolean isStartEnd(String activity) {
		return isStart(activity) || isEnd(activity);
	}
	private boolean isNode(String activity) {
		return isStartEnd(activity) || isActivity(activity) || isGateway(activity);
	}
	private boolean isEdge(String activity) {
		return activity.toUpperCase().endsWith("_EDGE");
	}
	private boolean isEdgeBendpoint(String activity) {
		return activity.toUpperCase().endsWith("_EDGE_BENDPOINT");
	}
	private boolean isReconnect(String activity) {
		return activity.toUpperCase().startsWith("RECONNECT_");
	}
	private boolean isEdgeLabel(String activity) {
		return activity.toUpperCase().endsWith("_EDGE_LABEL");
	}
	private boolean isEventOfInterest(String activity) {
		return isActivity(activity) || isGateway(activity) || isStartEnd(activity) || isEdge(activity) || isEdgeBendpoint(activity); 
	}
	private String getID(XEvent event) {
		try {
			return event.getAttributes().get("id").toString();
		} catch (Exception ex) {
			return "";
		}
	}
}